import logging

from lkauto.preprocessing.pruning import min_ratings_per_user, max_ratings_per_user
from lenskit.data import Dataset
from lenskit.data import from_interactions_df


def preprocess_data(data: Dataset,
                    user_col: str,
                    item_col: str,
                    rating_col: str = None,
                    timestamp_col: str = None,
                    include_timestamp: bool = True,
                    drop_na_values: bool = True,
                    drop_duplicates: bool = True,
                    min_interactions_per_user: int = None,
                    max_interactions_per_user: int = None) -> Dataset:
    """Preprocess data for LensKit
    This method can perform the following steps based on the user input:
    1. rename columns to "user", "item", "rating", "timestamp"
    2. Drop all rows with NaN values
    3. Drop all duplicate rows
    4. Drop all users with less than min_interactions_per_user interactions
    5. Drop all users with more than max_interactions_per_user interactions

    Parameters
    ----------
    data: pd.DataFrame
        Dataframe with columns "user", "item", "rating"
    user_col: str
        Name of the user column
    item_col: str
        Name of the item column
    rating_col: str
        Name of the rating column
    timestamp_col: str
        Name of the timestamp column
    include_timestamp: bool = True
        If True, the timestamp column will be included in the dataset
    drop_na_values: bool = True
        If True, all rows with NaN values will be dropped
    drop_duplicates: bool = True
        If True, all duplicate rows will be dropped
    min_interactions_per_user: int = None
        If not None, all users with less than this number of interactions will be dropped
    max_interactions_per_user: int = None
        If not None, all users with more than this number of interactions will be dropped

    Returns
    -------
    pd.DataFrame
        Dataframe with columns "user", "item", "rating"
    """

    logger = logging.getLogger('lenskit-auto')
    logger.info('--Start Preprocessing--')

    data = data.interaction_table(format='pandas')
    # original_cols = data.columns.tolist()
    # print(original_cols)

    # rename columns
    if include_timestamp:
        if rating_col is None:
            data = data[[user_col, item_col, timestamp_col]]
            data.columns = ['user', 'item', 'timestamp']
        else:
            # data = data[[user_col, item_col, rating_col, timestamp_col]]
            data.columns = ['user', 'item', 'rating', 'timestamp']
    else:
        if rating_col is None:
            data = data[[user_col, item_col]]
            data.columns = ['user', 'item']
        else:
            data = data[[user_col, item_col, rating_col]]
            data.columns = ['user', 'item', 'rating']

    # drop rows with NaN values
    if drop_na_values:
        logger.debug('Dropping rows with NaN values...')
        data = data.dropna()

    # drop duplicate rows
    if drop_duplicates:
        logger.debug('Dropping duplicate rows...')
        data = data.drop_duplicates(keep='first', inplace=False)

    # drop users with less than min_interactions_per_user interactions
    if min_interactions_per_user is not None:
        logger.debug('Dropping users with less than {} interactions...'.format(min_interactions_per_user))
        data = min_ratings_per_user(data, min_interactions_per_user)

    # drop users with more than max_interactions_per_user interactions
    if max_interactions_per_user is not None:
        logger.debug('Dropping users with more than {} interactions...'.format(max_interactions_per_user))
        data = max_ratings_per_user(data, max_interactions_per_user)

    logger.info('--End Preprocessing--')

    return from_interactions_df(data)



import logging

import numpy as np
import pandas as pd

from lenskit.data import Dataset, ItemListCollection
from lenskit.pipeline import predict_pipeline
from lenskit.batch import predict
from lenskit.metrics import RunAnalysis
from ConfigSpace import ConfigurationSpace

from lkauto.utils.filer import Filer
from lkauto.utils.get_model_from_cs import get_model_from_cs
from lkauto.utils.validation_split import validation_split
from lkauto.utils.update_top_n_runs import update_top_n_runs


class ExplicitEvaler:
    """ExplicitEvaler

        the ExplicitEvaler class handles the evaluation of rating prediction models.
        An Evaluation run consists of training a model and to predict and evaluate
        the performance on a validation split.

        Attributes
        ----------
        train : Dataset
            lenskit dataset containing the train split.
        optimization_metric: function
            LensKit prediction accuracy metric used to evaluate the model (either rmse or mae)
        filer : Filer
            Filer to organize the output.
        validation : ItemListCollection
            An ItemListCollection containing the validation split.
        random_state :
            The random number generator or seed (see :py:func:`lenskit.util.rng`).
        split_folds :
            The number of folds of the validation split
        split_strategy :
            The strategy used to split the data (either "user_based" or "row_based")
        split_frac :
            The fraction of the data used for the validation split. If the split_folds value is greater than 1,
            this value is ignored.
        ensemble_size :
            The number of models used to build the final ensemble predictor.
        minimize_error_metric_val :
            If True, the error metric is minimized. If False, the error metric is maximized. This parameter needs to be
            set in corelation with the optimization metric.

        Methods
        ----------
        evaluate_explicit(config_space: ConfigurationSpace) -> float
    """

    def __init__(self,
                 train: Dataset,
                 optimization_metric,
                 filer: Filer,
                 validation: ItemListCollection = None,
                 random_state=42,
                 split_folds: int = 1,
                 split_strategy: str = 'user_based',
                 split_frac: float = 0.25,
                 ensemble_size: int = 50,
                 minimize_error_metric_val: bool = True,
                 ) -> None:
        self.logger = logging.getLogger('lenskit-auto')
        self.train = train
        self.filer = filer
        self.validation = validation
        self.random_state = random_state
        self.split_folds = split_folds
        self.optimization_metric = optimization_metric
        self.split_strategy = split_strategy
        self.split_frac = split_frac
        self.minimize_error_metric_val = minimize_error_metric_val
        self.run_id = 0
        self.ensemble_size = ensemble_size
        self.top_n_runs = pd.DataFrame(columns=['run_id', 'model', 'error'])
        if self.validation is None:
            self.train_test_splits = validation_split(data=self.train,
                                                      strategy=self.split_strategy,
                                                      num_folds=self.split_folds,
                                                      frac=self.split_frac,
                                                      random_state=self.random_state)
        else:
            self.train_test_splits = None

    def evaluate(self, config_space: ConfigurationSpace) -> float:
        """ evaluates model defined in config_space

            The config_space parameter defines a model.
            This model is build, trained and evaluated with the validation split.

            Parameters
            ----------
            config_space : ConfigurationSpace
                configuration space containing information to build a model

            Returns
            ----------
            validation_error : float
                the error of the considered model
        """
        output_path = 'eval_runs/'
        self.run_id += 1
        error_metric = np.array([])
        validation_data = pd.DataFrame()

        # get model from configuration space
        model = get_model_from_cs(config_space, feedback='explicit')

        if self.validation is None:
            for fold in self.train_test_splits:
                validation_train = self.train_test_splits.train
                validation_test = self.train_test_splits.test

                pipeline = predict_pipeline(scorer=model)
                fit_pipeline = pipeline.clone()
                fit_pipeline.train(data=validation_train)

                recs = predict(fit_pipeline, validation_test)

                run_analysis = RunAnalysis()
                run_analysis.add_metric(self.optimization_metric)
                error_results = run_analysis.measure(recs, validation_test)

                error_metric = np.append(error_metric, error_results)
                validation_data = pd.concat([validation_data, recs], ignore_index=True)
        else:
            for fold in range(self.split_folds):
                validation_train = self.train
                validation_test = self.validation

                pipeline = predict_pipeline(scorer=model)
                fit_pipeline = pipeline.clone()
                fit_pipeline.train(data=validation_train)

                recs = predict(fit_pipeline, validation_test)

                run_analysis = RunAnalysis()
                run_analysis.add_metric(self.optimization_metric)
                error_results = run_analysis.measure(recs, validation_test)

                error_metric = np.append(error_metric, error_results)
                validation_data = pd.concat([validation_data, recs], ignore_index=True)

        # Save validation data for reproducibility and ensembling
        self.top_n_runs = update_top_n_runs(config_space=config_space,
                                            errors=error_metric,
                                            num_models=self.ensemble_size,
                                            run_id=self.run_id,
                                            top_n_runs=self.top_n_runs)
        self.filer.save_validataion_data(config_space=config_space,
                                         predictions=validation_data,
                                         metric_scores=error_metric,
                                         output_path=output_path,
                                         run_id=self.run_id)

        # calculate mean error_metric
        validation_error = error_metric.mean()

        self.logger.info('Run ID: ' + str(self.run_id) + ' | ' + str(config_space.get('algo')) + ' | ' +
                         self.optimization_metric.__name__ + ': ' + str(validation_error))
        self.logger.debug(str(config_space))

        # return error_metric
        if self.minimize_error_metric_val:
            return validation_error
        else:
            return 1 - validation_error



from lenskit.als import BiasedMFScorer, ImplicitMFScorer
from ConfigSpace import ConfigurationSpace, UniformIntegerHyperparameter, UniformFloatHyperparameter, CategoricalHyperparameter


class ImplicitMF(ImplicitMFScorer):
    def __init__(self, features, feedback="implicit", **kwargs):
        super().__init__(feedback=feedback, features=features, **kwargs)
        self.features = features  # store the features as an instance variable for testing

    @staticmethod
    def get_default_configspace(**kwargs):
        """
              return default configurationspace
        """
        features = UniformIntegerHyperparameter('features', lower=5, upper=10000, default_value=1000, log=True)
        ureg = UniformFloatHyperparameter('ureg', lower=0.01, upper=0.1, default_value=0.1, log=True)
        ireg = UniformFloatHyperparameter('ireg', lower=0.01, upper=0.1, default_value=0.1, log=True)

        cs = ConfigurationSpace()
        cs.add([features, ureg, ireg])
        return cs


class BiasedMF(BiasedMFScorer):
    def __init__(self, features, feedback="explicit", **kwargs):
        super().__init__(features=features, **kwargs)
        self.feedback = feedback
        self.features = features  # store the features as an instance variable for testing

    @staticmethod
    def get_default_configspace(**kwargs):
        """
               return default configuration spaces for hyperparameter
        """

        """
        The authors of the original ALS paper (https://link.springer.com/chapter/10.1007/978-3-540-68880-8_32) stated:
        The most important discovery we made is that ALS-WR never overfits the data if we either increase
        the number of iterations or the number of hidden features.
        The paper stated that the improvement of the performance maximized around 1000 features.
        Therefore, we will set the upper bound and the default value of features to 10000.
        Since the authors just evaluated on one larger dataset, we still allow smaller and larger feature numbers
        but set the default value to 1000.
        """
        # features = Integer('features', bounds=(2, 10000), default=1000, log=True)  # No default value given
        # no default value given but we set the default value to 1000???
        features = UniformIntegerHyperparameter('features', lower=2, upper=10000, default_value=1000, log=True)
        """
        The authors of the original ALS paper set the range of the regularization hyperparameter to from 0.03 - 0.065.
        https://link.springer.com/chapter/10.1007/978-3-540-68880-8_32
        Therefore we set the lower bound of the two regularization parameters (ureg and ireg) to 0.065.
        LensKit sets the default regularization hyperparameter to 0.1 Therefore we set the upper bound of the two
        regularization parameters (ureg and ireg) to 0.1.
        """
        ureg = UniformFloatHyperparameter('ureg', lower=0.01, upper=0.1, default_value=0.1, log=True)
        ireg = UniformFloatHyperparameter('ireg', lower=0.01, upper=0.1, default_value=0.1, log=True)

        """
        The damping hyperparameter en- or disables a damping factor.
        In the future we may want to tune the damping values as well.
        """
        bias = CategoricalHyperparameter('bias', choices=[True, False], default_value=True)

        cs = ConfigurationSpace()
        cs.add([features, ureg, ireg, bias])
        return cs



import pandas as pd
from ConfigSpace import ConfigurationSpace


def update_top_n_runs(num_models: int, top_n_runs: pd.DataFrame, run_id: int, config_space: ConfigurationSpace, errors):
    """
    updates the top n runs dataframe with the new run

    Parameters
    ----------
    num_models : int
        number of models to keep track of
    top_n_runs : pd.DataFrame
        pandas dataframe containing the top n runs of the optimization method.
    run_id : int
        run id of the new run
    config_space : ConfigurationSpace
        configuration space of the new run
    errors : np.ndarray
        errors of the new run

    Returns
    -------
    top_n_runs : pd.DataFrame
        pandas dataframe containing the top n runs of the optimization method.
    """
    # create a dataframe containing the run id, model and error of the new run
    model_performance = pd.DataFrame(data={'run_id': [run_id],
                                           'model': [config_space['algo']],
                                           'error': [errors.mean()]})

    # if the top n runs dataframe does not contain an entry for each model, add the new run to the dataframe
    if len(top_n_runs) < num_models:
        top_n_runs = pd.concat([top_n_runs, model_performance])
    # if the top n runs dataframe contains an entry for each model, check if the new run is better than the worst run
    else:
        # get the runs for each model type
        item_item_silo = top_n_runs[top_n_runs['model'] == 'ItemItem']
        user_user_silo = top_n_runs[top_n_runs['model'] == 'UserUser']
        als_bias_mf_silo = top_n_runs[top_n_runs['model'] == 'ALSBiasedMF']
        bias_silo = top_n_runs[top_n_runs['model'] == 'Bias']
        funk_svd_silo = top_n_runs[top_n_runs['model'] == 'FunkSVD']
        biased_svd_silo = top_n_runs[top_n_runs['model'] == 'BiasedSVD']

        # create a list containing the runs for each model type
        silo_list = [item_item_silo, user_user_silo, als_bias_mf_silo, bias_silo, funk_svd_silo, biased_svd_silo]

        max_len = 0
        max_index = 0

        # get the model type with the most runs
        for i, silo in enumerate(silo_list):
            if len(silo) > max_len:
                max_len = len(silo)
                max_index = i

        # get the runs for the model type with the most runs
        silo = top_n_runs[top_n_runs['model'] == config_space['algo']]

        if len(silo) == max_len:
            # get the worst run of the model type with the most runs
            worst_performance = silo['error'].max()

            # remove the worst run of the model type with the most runs
            top_n_runs = top_n_runs[top_n_runs['model'] != config_space['algo']]

            # add the new run to the dataframe
            silo = silo[silo['error'] < worst_performance]
            silo = pd.concat([silo, model_performance])

            # add the runs of the model type with the most runs to the dataframe
            top_n_runs = pd.concat([top_n_runs, silo])
        else:
            # get the worst run of the model type with the most runs
            max_silo = silo_list[max_index]

            # remove the worst run of the model type with the most runs
            worst_performance = max_silo['error'].max()
            top_n_runs = top_n_runs[top_n_runs['model'] != max_silo['model'].iloc[0]]
            max_silo = max_silo[max_silo['error'] < worst_performance]
            top_n_runs = pd.concat([top_n_runs, max_silo])

            # add the new run to the dataframe
            top_n_runs = pd.concat([top_n_runs, model_performance])

    return top_n_runs



import unittest

from ConfigSpace import Categorical
from ConfigSpace import ConfigurationSpace

from lkauto.utils.get_default_configurations import get_default_configurations


class TestGetDefaultConfigurations(unittest.TestCase):

    def setUp(self):
        self.algorithm_list = ['algo1', 'algo2', 'algo3']
        self.config_space = ConfigurationSpace(
            space={
                "algo": Categorical("algo", self.algorithm_list, default="algo1"),
            }
        )
        for algorithm in self.algorithm_list:
            self.config_space.add_configuration_space(
                prefix=algorithm,
                delimiter=":",
                configuration_space=ConfigurationSpace(),
                parent_hyperparameter={"parent": self.config_space["algo"], "value": algorithm},
            )

        print(self.config_space)

    def test_getDefaultConfigurations_givenValidConfigurationSpace_correctConfigurationSpaceListReturnedExpected(self):
        result = get_default_configurations(self.config_space)

        for i, algorithm in enumerate(self.algorithm_list):
            self.assertEqual(algorithm, result[i].get("algo"))


if __name__ == '__main__':
    unittest.main()



import unittest
import uuid
from unittest.mock import MagicMock, patch

from lenskit.als import ImplicitMFScorer, BiasedMFScorer
from lenskit.basic import BiasScorer
from lenskit.funksvd import FunkSVDScorer
from lenskit.knn import UserKNNScorer, ItemKNNScorer
from lenskit.sklearn.svd import BiasedSVDScorer

from lkauto.utils.get_model_from_cs import get_model_from_cs


class TestGetModelFromCS(unittest.TestCase):

    def setUp(self):
        self.cs = MagicMock()
        self.fallback_model = MagicMock()
        self.random_state = 42

    def test_getModelFromCS_givenInvalidFeedback_valueErrorThrown(self):
        with self.assertRaises(ValueError) as cm:
            get_model_from_cs(cs=self.cs, random_state=self.random_state,
                              feedback="")
        self.assertEqual("Unknown feedback type: ", cm.exception.args[0])

    def test_getModelFromCS_givenInvalidAlgorithmInCS_valueErrorThrown(self):
        self.cs.get.return_value = 'alg'
        self.cs.items.return_value = [('algo', 'alg'), ('alg:attr1', 'val1'), ('alg:attr2', 'val2')]

        with self.assertRaises(ValueError) as cm:
            get_model_from_cs(cs=self.cs, random_state=self.random_state,
                              feedback="implicit")
        self.assertEqual("Unknown algorithm: alg", cm.exception.args[0])

    def test_getModelFromCS_givenImplicitAndValidInputs_correctModelReturnedExpected(self):
        algorithm_params_list = [
            ('ItemItem', [('algo', 'ItemItem'), ('ItemItem:attr1', 'val1')], ItemKNNScorer),
            ('UserUser', [('algo', 'UserUser'), ('UserUser:attr1', 'val1')], UserKNNScorer),
            ('ImplicitMF',[('algo', 'ImplicitMF'), ('ImplicitMF:ureg', 0.15), ('ImplicitMFScorer:ireg', 0.15)], ImplicitMFScorer)]

        for algorithm_params in algorithm_params_list:
            with self.subTest(algorithm=algorithm_params[0]):
                self.cs.get.return_value = algorithm_params[0]
                self.cs.items.return_value = algorithm_params[1]
                algorithm_mock = MagicMock()
                algorithm_mock_test_id = uuid.uuid4()
                algorithm_mock.test_id = algorithm_mock_test_id


                result = get_model_from_cs(cs=self.cs, random_state=self.random_state, feedback="implicit")

                if algorithm_params[0] != "ImplicitMF":
                    self.assertEqual("implicit", result.config.feedback)
                self.assertEqual(algorithm_params[-1], type(result))

    def test_getModelFromCS_givenExplicitAndValidInputs_correctModelReturnedExpected(self):
        algorithm_params_list = [
            ('ItemItem', [('algo', 'ItemItem'), ('ItemItem:attr1', 'val1')], ItemKNNScorer),
            ('UserUser', [('algo', 'UserUser'), ('UserUser:attr1', 'val1')], UserKNNScorer),
            ('FunkSVD', [('algo', 'FunkSVD'), ('FunkSVD:attr1', 'val1')], FunkSVDScorer),
            ('BiasedSVD', [('algo', 'BiasedSVD'), ('BiasedSVD:attr1', 'val1')], BiasedSVDScorer),
            ('ALSBiasedMF', [('algo', 'ALSBiasedMF'), ('ALSBiasedMF:ureg', 1), ('BiasedMFScorer:ireg', 2)], BiasedMFScorer),
            ('Bias', [('algo', 'Bias'), ('Bias:user_damping', 'val1'), ('BiasScorer:item_damping', 'val2')], BiasScorer)
        ]

        for algorithm_params in algorithm_params_list:
            with self.subTest(algorithm=algorithm_params[0]):
                self.cs.get.return_value = algorithm_params[0]
                self.cs.items.return_value = algorithm_params[1]
                algorithm_mock = MagicMock()
                algorithm_mock_test_id = uuid.uuid4()
                algorithm_mock.test_id = algorithm_mock_test_id

                result = get_model_from_cs(cs=self.cs, random_state=self.random_state, feedback="explicit")

                self.assertEqual(algorithm_params[-1], type(result))

                '''
                with patch('lkauto.utils.get_model_from_cs.{}'.format(
                    algorithm_params[0] if algorithm_params[0] != 'ALSBiasedMF' else 'BiasedMF'),
                    return_value=algorithm_mock):
                    result = get_model_from_cs(cs=self.cs,
                                               random_state=self.random_state,
                                               feedback="explicit")

                self.assertEqual(algorithm_mock_test_id, result.test_id)
                '''


if __name__ == '__main__':
    unittest.main()



import pandas as pd
import numpy as np
from lenskit.data import Dataset

from lkauto.utils.filer import Filer
from lkauto.utils.get_model_from_cs import get_model_from_cs

from lkauto.ensemble.greedy_ensemble_selection import EnsembleSelection


def build_ensemble(train: Dataset,
                   top_n_runs: pd.DataFrame,
                   filer: Filer,
                   ensemble_size: int,
                   lenskit_metric,
                   maximize_metric: bool):
    config_ids = top_n_runs.sort_values(by='error', ascending=True)['run_id']
    ensemble_y = train.interaction_table(format="pandas")['rating']
    ensemble_x = []
    val_indices = None
    bm_cs_list = []

    for config_id in config_ids:
        # Load predictions file for config id, sort and append to ensemble_X
        bm_pred = filer.get_dataframe_from_csv(path_to_file='eval_runs/{}/predictions.csv'.format(config_id))
        bm_cs = filer.get_dict_from_json_file(path_to_file='eval_runs/{}/config_space.json'.format(config_id))

        bm_cs_list.append(bm_cs)

        # Get Validation indices (overwrite for now but should be the same for each base model)
        bm_pred = bm_pred.sort_values(by=list(bm_pred)[0])
        if (val_indices is not None) and (not val_indices.equals(bm_pred[list(bm_pred)[0]])):
            raise ValueError("Validation Indices are not identical between base models!")
        val_indices = bm_pred[list(bm_pred)[0]]

        # Append predictions to ensemble train X
        ensemble_x.append(np.array(bm_pred[list(bm_pred)[1]]))

    ensemble_y = np.array(ensemble_y.loc[val_indices])

    es = EnsembleSelection(ensemble_size=ensemble_size, lenskit_metric=lenskit_metric, maximize_metric=maximize_metric)
    es.ensemble_fit(ensemble_x, ensemble_y)
    es.base_models = [get_model_from_cs(cs, feedback='explicit') for cs, weight in zip(bm_cs_list, es.weights_) if weight > 0]
    es.old_to_new_idx = {old_i: new_i for new_i, old_i in enumerate([idx for idx, weight in enumerate(es.weights_) if weight > 0])}

    incumbent = {"model": str(es),
                 "top_50_models": list(bm_cs_list),
                 "trajectory": list(es.trajectory_),
                 "ensamble_size": es.ensemble_size,
                 "train_loss": es.train_loss_,
                 "weights": list(es.weights_)}

    return es, incumbent



import unittest

import numpy as np
import pandas as pd

from lkauto.utils.validation_split import validation_split
from lenskit.data import from_interactions_df


class TestValidationSplit(unittest.TestCase):

    def setUp(self):
        self.df = pd.DataFrame(np.array([[1, 1, 1],
                                         [1, 2, 2],
                                         [1, 3, 3],
                                         [2, 1, 1],
                                         [2, 2, 2],
                                         [2, 3, 3],
                                         [3, 1, 1],
                                         [3, 2, 2],
                                         [3, 3, 3],
                                         [4, 1, 1],
                                         [4, 2, 2],
                                         [4, 3, 3],
                                         [5, 1, 1],
                                         [5, 2, 2],
                                         [5, 3, 3]]), columns=["user", "item", "rating", ])
        self.ds = from_interactions_df(self.df)

    """
    def test_validationSplit_givenValidDataFrame_correctSplitTrainAndValidationDataframesReturnedExpected(self):
        val_fold_indices = validation_split(data=self.ds, frac=0.25, random_state=42)

        validation_train = self.df.loc[val_fold_indices[0]["train"], :]
        validation_test = self.df.loc[val_fold_indices[0]["validation"], :]

        self.assertTrue(validation_train.shape == (75, 3))
        self.assertTrue(validation_test.shape == (25, 3))
        """

    def test_validationSplit_givenValidDataset_1Fold_UserBased(self):
        splits = validation_split(data=self.ds, strategy="user_based", frac=0.2, num_folds=1, random_state=42)

        fold = next(splits)
        test_sample_fold = fold.test
        train_sample_fold = fold.train

        self.assertTrue(test_sample_fold.to_df().shape[0] == 3)
        self.assertTrue(train_sample_fold.interaction_count == 12)

    def test_validationSplit_givenValidDataset_3Fold_RowBased(self):
        splits = validation_split(data=self.ds, strategy="row_based", frac=0.2, num_folds=3, random_state=42)

        fold = next(splits)
        test_sample_fold = fold.test
        train_sample_fold = fold.train

        self.assertTrue(test_sample_fold.to_df().shape[0] == 5)
        self.assertTrue(train_sample_fold.interaction_count == 10)




if __name__ == '__main__':
    unittest.main()



from lenskit.knn.user import UserKNNScorer, UserKNNConfig
from ConfigSpace import UniformIntegerHyperparameter, UniformFloatHyperparameter
from ConfigSpace import ConfigurationSpace


class UserUser(UserKNNScorer):
    def __init__(self, max_nbrs, min_nbrs=1, min_sim=1e-6, feedback="implicit", **kwargs):
        # store the parameters as an instance variables so we can acces it (for testing)
        self.feedback = feedback
        self.max_nbrs = max_nbrs
        self.min_nbrs = min_nbrs
        self.min_sim = min_sim

        config = UserKNNConfig(
            feedback=feedback,
            max_nbrs=max_nbrs,
            min_nbrs=min_nbrs,
            min_sim=min_sim,
            **kwargs
        )
        super().__init__(config=config, **kwargs)

    @staticmethod
    def get_default_configspace(**kwargs):
        """
            return default configuration spaces for hyperparameter
        """
        """
        The max_nbrs hyperparameter is set to 10000 in LensKit-Auto. Generally speaking, the higher the max_nbrs
        hyperparameter value, the better the performance. But the computational cost will increase
        exponentially by increasing the max_nbrs hyperparameter value. 10000 is a reasonable value for max_nbrs
        hyperparameter since it has relatively good performance and is still able
        to run in a reasonable amount of time.
        """
        max_nbrs = UniformIntegerHyperparameter('max_nbrs', lower=1, upper=10000, default_value=1000, log=True)

        """
        The min_nbrs hyperparameter describes the minimum number of neighbors for scoring each item.
        Since the LensKit default value for the min_nbrs hyperparameter is 1, we set the lower bound  to 1.
        The upper bound is set to the max_nbrs hyperparameter value.
        Therefore, the upper bound of min_nbrs is set to 10000 to cover the full possible range of the
        min_nbrs hyperparameter.
        """
        min_nbrs = UniformIntegerHyperparameter('min_nbrs', lower=1, upper=1000, default_value=1, log=True)

        """
        The min_sim hyperparameter describes the minimum threshold for similarity between items. It is commonly
        refered as the minimum support constraint. The min_sim hyperparameter limits the number of items that are taken
        into account for the similarity calculation.
        The following constrains are taken from :cite:t:`Deshpande2004-ht`
        A high value will result in a higher-order scheme that uses
        very few itemsets and as such it does not utilize its full potential, whereas a low value may lead to an
        exponentially large number of itemsets, making it computationally intractable.
        Unfortunately, there are no good ways to a priori select the value of support. This is because for a
        given value of σ the number of frequent item sets that exist in a dataset depends on the dataset’s density
        and the item co-occurrence patterns in the various rows.
        Since the paper already states that it is very difficult to find the best value, we define a large bound around
        the default LensKit value.
        """
        min_sim = UniformFloatHyperparameter('min_sim', lower=1.0e-10, upper=1.0e-2, default_value=1.0e-6, log=True)

        cs = ConfigurationSpace()
        cs.add([max_nbrs, min_nbrs, min_sim])

        return cs



import unittest
from unittest.mock import patch, MagicMock

import numpy as np
import pandas as pd

from ConfigSpace import ConfigurationSpace
from lenskit.data import from_interactions_df, DatasetBuilder

from lkauto.utils.get_default_configuration_space import get_default_configuration_space
from lkauto.utils.validation_split import validation_split


class TestGetDefaultConfigurationSpace(unittest.TestCase):
    model_mock = MagicMock()
    model_mock.get_default_configspace.return_value = ConfigurationSpace()

    def setUp(self):
        self.random_state = 42
        self.df = pd.DataFrame(np.array([[1,1,1],
                            [1,2,2],
                            [1,3,3],
                            [2,1,1],
                            [2,2,2],
                            [2,3,3],
                            [3,1,1],
                            [3,2,2],
                            [3,3,3],
                            [4,1,1],
                            [4,2,2],
                            [4,3,3],
                            [5,1,1],
                            [5,2,2],
                            [5,3,3]]), columns=["user", "item", "rating", ])
        # self.df = pd.DataFrame(np.ones((100, 3), dtype=int), columns=["user", "item", "rating", ])
        self.ds = from_interactions_df(self.df, user_col="user", item_col="item", rating_col="rating")
        self.val_fold_indices = validation_split(data=self.ds, random_state=42)

    def test_getDefaultConfigurationSpace_givenInvalidFeedback_valueErrorThrown(self):
        with self.assertRaises(ValueError) as cm:
            get_default_configuration_space(feedback="", data=self.ds,
                                            val_fold_indices=self.val_fold_indices,
                                            random_state=self.random_state)
        self.assertEqual("Unknown feedback type: ", cm.exception.args[0])

    @patch('lkauto.utils.get_default_configuration_space.ItemItem', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.UserUser', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.FunkSVD', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.BiasedSVD', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.ImplicitMF', model_mock)
    def test_getDefaultConfigurationSpace_GivenImplicitAndValidInputs_CorrectConfigSpaceReturnedExpected(self):
        algorithm_list_string = 'ItemItem, UserUser, ImplicitMF'

        result = get_default_configuration_space(feedback='implicit', data=self.ds,
                                                 val_fold_indices=self.val_fold_indices,
                                                 random_state=self.random_state)

        self.assertIsInstance(result, ConfigurationSpace)
        self.assertTrue(algorithm_list_string in str(result.get('algo')))

    @patch('lkauto.utils.get_default_configuration_space.ItemItem', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.UserUser', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.FunkSVD', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.BiasedSVD', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.BiasedMF', model_mock)
    @patch('lkauto.utils.get_default_configuration_space.Bias', model_mock)
    def test_getDefaultConfigurationSpace_GivenExplicitAndValidInputs_CorrectConfigSpaceReturnedExpected(self):
        algorithm_list_string = 'ItemItem, UserUser, FunkSVD, BiasedSVD, ALSBiasedMF, Bias'

        result = get_default_configuration_space(feedback='explicit', data=self.ds,
                                                 val_fold_indices=self.val_fold_indices,
                                                 random_state=self.random_state)

        self.assertIsInstance(result, ConfigurationSpace)
        self.assertTrue(algorithm_list_string in str(result.get('algo')))


if __name__ == '__main__':
    unittest.main()



from ConfigSpace import ConfigurationSpace
from ConfigSpace import Configuration
from ConfigSpace import Constant
from typing import List


def get_default_configurations(config_space: ConfigurationSpace) -> List[Configuration]:
    """
    returns a list of default configurations for all algorithms in the configuration space

    Parameters
    ----------
    config_space : ConfigurationSpace
        configuration space to use

    Returns
    -------
    List[Configuration]
        List of default configurations for all algorithms in the configuration space
    """

    # get all algorithms in the configuration space
    if type(config_space.get('algo')) == Constant:
        # algorithms = [config_space.get('algo').value]
        algorithms = [config_space['algo'].value]
    else:
        # algorithms = config_space.get('algo').choices
        algorithms = config_space["algo"].choices
    # initialize a list for default configurations for each algorithm
    initial_configuration_list = []

    # get the default configuration for each algorithm and store it in the list
    for algorithm in algorithms:
        # set the default value for the algorithm hyperparameter
        config_space.get('algo').default_value = str(algorithm)
        # get the default configuration for the algorithm
        configuration = config_space.get_default_configuration()
        # add the configuration to the list
        initial_configuration_list.append(configuration)

    # return the list of default configurations
    return initial_configuration_list



from lenskit.data import Dataset, from_interactions_df


def min_ratings_per_user(dataset: Dataset, num_ratings: int, count_duplicates: bool = False):
    """Prune users with less than num_ratings ratings

    Parameters
    ----------
    dataset: Dataset
        LensKit Dataset object containing user-item interactions with ratings
    num_ratings: int
        Minimum number of ratings per user
    count_duplicates: bool = False
        If True, all ratings are counted, otherwise only unique ratings are counted

    Returns
    -------
    Dataset
        Filtered Dataset with only users meeting the minimum rating threshold
        the Dataset will contain the columns "user_id", "item_id", "rating"
    """
    # get the user statistics from the dataset
    user_stats = dataset.user_stats()
    if count_duplicates:
        valid_users = user_stats[user_stats['count'] >= num_ratings].index  # count: total number of ratings (including duplicates)
    else:
        valid_users = user_stats[user_stats['item_count'] >= num_ratings].index  # item_count: number of unique items rated
    # convert the interaction table to a pandas DataFrame and filter by valid users
    users_of_interest = dataset.interaction_table(format='pandas', original_ids=True)
    users_of_interest = users_of_interest[users_of_interest['user_id'].isin(valid_users)]
    return from_interactions_df(users_of_interest)


def max_ratings_per_user(dataset: Dataset, num_ratings: int, count_duplicates: bool = False):
    """Prune users with more than num_ratings ratings

    Parameters
    ----------
    dataset: Dataset
        LensKit Dataset object containing user-item interactions with ratings
    num_ratings: int
        Maximum number of ratings per user
    count_duplicates: bool = False
        If True, all ratings are counted, otherwise only unique ratings are counted

    Returns
    -------
    Dataset
        Filtered Dataset with only users meeting the minimum rating threshold
        the Dataset will contain the columns "user_id", "item_id", "rating"
    """

    user_stats = dataset.user_stats()
    if count_duplicates:
        valid_users = user_stats[user_stats['count'] <= num_ratings].index  # count: total number of ratings (including duplicates)
    else:
        valid_users = user_stats[user_stats['item_count'] <= num_ratings].index  # item_count: number of unique items rated
    # convert the interaction table to a pandas DataFrame and filter by valid users
    users_of_interest = dataset.interaction_table(format='pandas', original_ids=True)
    users_of_interest = users_of_interest[users_of_interest['user_id'].isin(valid_users)]
    return from_interactions_df(users_of_interest)



import unittest

import ConfigSpace as CS

from lkauto.algorithms.svd import BiasedSVD


class TestSVD(unittest.TestCase):
    def test_init_givenObjectInitialized_ObjectInitializedCorrectlyExpected(self):
        expected_features = 10
        biased_svd = BiasedSVD(expected_features)
        self.assertIsInstance(biased_svd, BiasedSVD)
        self.assertEqual(expected_features, biased_svd.features)

    def test_defaultConfigspace_GivenFunctionCalled_correctConfigSpaceReturnedExpected(self):
        number_item_list = [1500, 100000]
        for number_item in number_item_list:
            cs = BiasedSVD.get_default_configspace(number_item=number_item)
            params = list(cs.values())
            with self.subTest(number_item=number_item):
                self.assertIsInstance(cs, CS.ConfigurationSpace)
                if number_item < 10000:
                    self.assertTrue(any(
                        param.name == "features" and param.default_value == 1499 and param.lower == 2
                        and param.upper == 1500 for param in params))
                else:
                    self.assertTrue(any(
                        param.name == "features" and param.default_value == 1000 and param.lower == 2
                        and param.upper == 10000 for param in params))
                self.assertTrue(any(
                    param.name == "damping" and param.default_value == 5 and param.lower == 0.0001
                    and param.upper == 1000 for param in params))


if __name__ == '__main__':
    unittest.main()



# Code Taken from here with adaptions to be usable:
# https://github.com/automl/auto-sklearn/blob/master/autosklearn/ensembles/ensemble_selection.py

from collections import Counter
from typing import List, Union

import numpy as np
import pandas as pd
from lenskit.data import Dataset


class EnsembleSelection:
    """An ensemble of selected algorithms

    Fitting an EnsembleSelection generates an ensemble from the the models
    generated during the search process. Can be further used for prediction.

    Parameters
    ----------
    lenskit_metric: metric from lenskit
        The metric used to evaluate the models
    maximize_metric: bool = False
        If the metric is to be optimized or not.
    """

    def __init__(self, ensemble_size: int, lenskit_metric,
                 maximize_metric: bool = False) -> None:

        self.ensemble_size = ensemble_size

        if maximize_metric:
            def minimized_metric(y_ture, y_pred):
                return -lenskit_metric(y_pred, y_ture)
        else:
            def minimized_metric(y_ture, y_pred):
                return lenskit_metric(y_pred, y_ture)

        self.metric = minimized_metric

        # Will be filled later from external
        self.base_models = None

    def fit(self, data: Dataset):
        """ Fit base models (we assume the ensemble part, ensemble_fit, was already fitted here or is fitted later)

        Parameters
        ----------
        data: Dataset
            Dataset with columns "user", "item", "rating"
        """
        if self.base_models is None:
            raise ValueError("Base Models is None; we need a list of base models to fit them here!")

        for bm in self.base_models:
            bm.fit(data)

        return self

    def predict(self, x_data: Dataset):
        """
        "user", "item" Dataframe
        """
        bm_preds = [bm.predict(x_data) for bm in self.base_models]
        test_ind = bm_preds[0].index
        ens_predictions = self.ensemble_predict([np.array(bm_pred) for bm_pred in bm_preds])

        return pd.Series(ens_predictions, index=test_ind)

    def ensemble_fit(self, predictions: List[np.ndarray], labels: np.ndarray):
        self.ensemble_size = int(self.ensemble_size)
        if self.ensemble_size < 1:
            raise ValueError('Ensemble size cannot be less than one!')

        self._fast(predictions, labels)
        self._apply_use_best()
        self._calculate_weights()

        return self

    def _fast(self, predictions: List[np.ndarray], labels: np.ndarray) -> None:
        """Fast version of Rich Caruana's ensemble selection method."""
        self.num_input_models_ = len(predictions)

        ensemble = []  # type: List[np.ndarray]
        trajectory = []
        order = []

        ensemble_size = self.ensemble_size

        weighted_ensemble_prediction = np.zeros(
            predictions[0].shape,
            dtype=np.float64,
        )
        fant_ensemble_prediction = np.zeros(
            weighted_ensemble_prediction.shape,
            dtype=np.float64,
        )
        for i in range(ensemble_size):
            losses = np.zeros(
                (len(predictions)),
                dtype=np.float64,
            )
            s = len(ensemble)
            if s > 0:
                np.add(
                    weighted_ensemble_prediction,
                    ensemble[-1],
                    out=weighted_ensemble_prediction,
                )

            # Memory-efficient averaging!
            for j, pred in enumerate(predictions):
                # fant_ensemble_prediction is the prediction of the current ensemble
                # and should be ([predictions[selected_prev_iterations] + predictions[j])/(s+1)
                # We overwrite the contents of fant_ensemble_prediction
                # directly with weighted_ensemble_prediction + new_prediction and then scale for avg
                np.add(
                    weighted_ensemble_prediction,
                    pred,
                    out=fant_ensemble_prediction
                )
                np.multiply(
                    fant_ensemble_prediction,
                    (1. / float(s + 1)),
                    out=fant_ensemble_prediction
                )

                losses[j] = self.metric(labels, fant_ensemble_prediction)

            all_best = np.argwhere(losses == np.nanmin(losses)).flatten()
            best = all_best[0]

            ensemble.append(predictions[best])
            trajectory.append(losses[best])
            order.append(best)

            # Handle special case
            if len(predictions) == 1:
                break

        self.indices_ = order
        self.trajectory_ = trajectory
        self.train_loss_ = trajectory[-1]

    def _apply_use_best(self):
        # Basically from autogluon the code
        min_score = np.min(self.trajectory_)
        idx_best = self.trajectory_.index(min_score)
        self.indices_ = self.indices_[:idx_best + 1]
        self.trajectory_ = self.trajectory_[:idx_best + 1]
        self.ensemble_size = idx_best + 1
        self.train_loss_ = self.trajectory_[idx_best]

    def _calculate_weights(self) -> None:
        ensemble_members = Counter(self.indices_).most_common()
        weights = np.zeros(
            (self.num_input_models_,),
            dtype=np.float64,
        )
        for ensemble_member in ensemble_members:
            weight = float(ensemble_member[1]) / self.ensemble_size
            weights[ensemble_member[0]] = weight

        if np.sum(weights) < 1:
            weights = weights / np.sum(weights)

        self.weights_ = weights

    def ensemble_predict(self, predictions: Union[np.ndarray, List[np.ndarray]]) -> np.ndarray:

        average = np.zeros_like(predictions[0], dtype=np.float64)
        tmp_predictions = np.empty_like(predictions[0], dtype=np.float64)

        # if predictions.shape[0] == len(self.weights_),
        # predictions include those of zero-weight models.
        if len(predictions) == len(self.weights_):
            for pred, weight in zip(predictions, self.weights_):
                np.multiply(pred, weight, out=tmp_predictions)
                np.add(average, tmp_predictions, out=average)

        # if prediction model.shape[0] == len(non_null_weights),
        # predictions do not include those of zero-weight models.
        elif len(predictions) == np.count_nonzero(self.weights_):
            non_null_weights = [w for w in self.weights_ if w > 0]
            for pred, weight in zip(predictions, non_null_weights):
                np.multiply(pred, weight, out=tmp_predictions)
                np.add(average, tmp_predictions, out=average)

        # If none of the above applies, then something must have gone wrong.
        else:
            raise ValueError("The dimensions of ensemble predictions"
                             " and ensemble weights do not match!")
        del tmp_predictions
        return average



import unittest

import ConfigSpace as CS

from lkauto.algorithms.item_knn import ItemItem


class TestItemKNN(unittest.TestCase):
    def test_init_givenObjectInitialized_ObjectInitializedCorrectlyExpected(self):
        expected_max_nbrs = 10
        item_item = ItemItem(expected_max_nbrs)
        self.assertEqual(expected_max_nbrs, item_item.max_nbrs)
        self.assertIsInstance(item_item, ItemItem)

    def test_defaultConfigspace_GivenFunctionCalled_correctConfigSpaceReturnedExpected(self):
        cs = ItemItem.get_default_configspace()
        params = list(cs.values())
        self.assertIsInstance(cs, CS.ConfigurationSpace)
        self.assertTrue(any(
            param.name == "max_nbrs" and param.default_value == 1000 and param.lower == 1 and param.upper == 10000 for
            param in params))
        self.assertTrue(any(
            param.name == "min_nbrs" and param.default_value == 1 and param.lower == 1 and param.upper == 1000 for
            param in params))
        self.assertTrue(any(
            param.name == "min_sim" and param.default_value == 1.0e-6 and param.lower == 1.0e-10 and param.upper ==
            1.0e-2 for param in params))


if __name__ == '__main__':
    unittest.main()



from pathlib import Path
from ConfigSpace import ConfigurationSpace
import pandas as pd
import numpy as np
import os
import json


class Filer:
    """Filer to handle the LensKit-Auto output

        This filer supports to structure the LensKit-Auto output in the file system

        Attributes
        ----------
        output_directory_path: path to the output directory
            path that leads to the output directory of LensKit-Auto

        Methods
        ----------
        get_output_directory_path() -> str
        get_smac_output_directory_path() -> str
        set_output_directory_path(output_directory_path: str) -> None
        save_dataframe_as_csv(self, dataframe: pd.DataFrame, output_path: str, name: str) -> None
        save_dictionary_to_json(self, dictionary: dict, output_path: str, name: str) -> None
        save_metric_scores_to_txt(self, metric_scores: np.array, output_path: str, name: str) -> None:
        get_dataframe_from_csv(self, path_to_file: str, index_column=None) -> pd.DataFrame:
        get_series_from_csv(self, path_to_file: str, index_column=None) -> pd.Series
        get_dict_from_json_file(self, path_to_file: str) -> dict
        get_numpy_array_from_txt_file(self, path_to_file: str) -> np.array
        append_dataframe_to_csv(self, dataframe: pd.DataFrame, output_path: str, name: str) -> None
        save_validataion_data(self,
                              config_space: ConfigurationSpace,
                              predictions: pd.DataFrame,
                              metric_scores: np.array,
                              output_path: str,
                              run_id: int) -> None
    """
    def __init__(self, output_directory_path='output/'):
        self.output_directory_path = output_directory_path
        Path(output_directory_path).mkdir(parents=True, exist_ok=True)

    def __check_if_folder_structure_exist(self, output_path: str) -> None:
        folder_path = os.path.join(self.output_directory_path, output_path)
        Path(folder_path).mkdir(parents=True, exist_ok=True)

    def get_output_directory_path(self) -> str:
        return self.output_directory_path

    def get_smac_output_directory_path(self) -> str:
        return os.path.join(self.output_directory_path, 'smac')

    def set_output_directory_path(self, output_directory_path: str) -> None:
        self.output_directory_path = output_directory_path

    def save_dataframe_as_csv(self, dataframe: pd.DataFrame, output_path: str, name: str) -> None:
        self.__check_if_folder_structure_exist(output_path=output_path)
        data_path = os.path.join(self.output_directory_path, output_path, '{}.csv'.format(name))
        dataframe.to_csv(path_or_buf=data_path)

    def save_dictionary_to_json(self, dictionary: dict, output_path: str, name: str) -> None:
        self.__check_if_folder_structure_exist(output_path)
        data_path = os.path.join(self.output_directory_path, output_path, '{}.json'.format(name))

        temp_dic = {}
        for key, value in dictionary.items():
            if isinstance(value, np.bool):
                if value:
                    temp_dic[key] = True
                else:
                    temp_dic[key] = False
            else:
                temp_dic[key] = value

        with open(data_path, 'w') as f:
            json.dump(temp_dic, f)

    def save_metric_scores_to_txt(self, metric_scores: np.array, output_path: str, name: str) -> None:
        self.__check_if_folder_structure_exist(output_path=output_path)
        data_path = os.path.join(self.get_output_directory_path(), output_path, '{}.txt'.format(name))

        np.savetxt(data_path, metric_scores)

    def get_dataframe_from_csv(self, path_to_file: str, index_column=None) -> pd.DataFrame:
        if index_column is None:
            dataframe = pd.read_csv(os.path.join(self.get_output_directory_path(), path_to_file))
        else:
            dataframe = pd.read_csv(os.path.join(self.get_output_directory_path(), path_to_file),
                                    index_col=index_column)
        return dataframe

    def get_series_from_csv(self, path_to_file: str, index_column=None) -> pd.Series:
        if index_column is None:
            series = pd.read_csv(os.path.join(self.get_output_directory_path(), path_to_file))
        else:
            series = pd.read_csv(os.path.join(self.get_output_directory_path(), path_to_file),
                                 index_col=index_column)
        series = series.iloc[:, 0]
        return series

    def get_dict_from_json_file(self, path_to_file: str) -> dict:
        with open(os.path.join(self.get_output_directory_path(), path_to_file)) as f:
            temp_dic = json.load(f)

        dictionary = {}

        for key, value in temp_dic.items():
            if isinstance(value, bool):
                if value:
                    dictionary[key] = True
                else:
                    dictionary[key] = False
            else:
                dictionary[key] = value

        return dictionary

    def get_numpy_array_from_txt_file(self, path_to_file: str) -> np.array:
        return np.loadtxt(fname=os.path.join(self.get_output_directory_path(), path_to_file))

    def append_dataframe_to_csv(self, dataframe: pd.DataFrame, output_path: str, name: str) -> None:
        self.__check_if_folder_structure_exist(output_path=output_path)
        data_path = os.path.join(self.get_output_directory_path(), output_path, '{}.csv'.format(name))

        dataframe.to_csv(data_path, mode='a', header=False, index=False)

    def save_validataion_data(self,
                              config_space: ConfigurationSpace,
                              predictions: pd.DataFrame,
                              metric_scores: np.array,
                              output_path: str,
                              run_id: int) -> None:
        """method to simplify the validation data output

                The validation data output is mainly used to build ensambles of the best models.
                Therefore, predictions, error scores and configuration spaces of each run need
                to be stored to the file system

                Parameters
                ----------
                config_space : ConfigurationSpace
                    configuraiton space of run
                predictions : pd.Dataframe
                    dataframe containing raw predictions
                metric_scores : np.array
                    numpy array containing metric values
                    Depending on the metric, differnet kind of metric values can be stored
                    in the metric scores array
                output_path : str
                    path to output folder
                run_id : int
                    id of smac search iteration
            """
        output_path = os.path.join(output_path, str(run_id))
        dictionary = dict(config_space)
        self.save_metric_scores_to_txt(metric_scores=metric_scores, output_path=output_path, name='rmse')
        self.save_dictionary_to_json(dictionary=dictionary, output_path=output_path, name='config_space')
        self.save_dataframe_as_csv(dataframe=predictions, output_path=output_path, name='predictions')



from lenskit.funksvd import FunkSVDScorer
from ConfigSpace import ConfigurationSpace, UniformIntegerHyperparameter, UniformFloatHyperparameter


class FunkSVD(FunkSVDScorer):
    def __init__(self, features, feedback="explicit", **kwargs):
        super().__init__(features=features, **kwargs)
        # store feature as an instance variable so we can acces it (for testing)
        self.feedback = feedback
        self.features = features

    @staticmethod
    def get_default_configspace(**kwargs):
        """
               return default configurationspace
        """
        features = UniformIntegerHyperparameter('features', lower=2, upper=10000, default_value=1000, log=True)

        """
        The authors of the original FunkSVD paper (https://sifter.org/~simon/journal/20061211.html) stated:
        Lrate is the learning rate, a rather arbitrary number which I fortuitously set to 0.001 on day one
        and regretted it every time I tried anything else after that. Err is the residual error from the
        current prediction.
        But the original dataset just evaluated the performance on the netflix price dataset. Other datasets
        perform well on ranges around 0.001.
        Therefore, the pip install -e .loatHyperparameter('lrate', lower=0.0001, upper=0.01, default_value=0.001)

        """
        lrate = UniformFloatHyperparameter('lrate', lower=0.0001, upper=0.01, default_value=0.001)
        """
        The authors of the original FunkSVD paper (https://sifter.org/~simon/journal/20061211.html) stated:
        The point here is to try to cut down on over fitting, ultimately allowing us to use
        more features. Last I recall, Vincent liked K=0.02 or so, with well over 100 features (singular vector
        pairs--if you can still call them that).
        The default value of 0.02 is considered for the range. The range is set to a close range around the 0.02 value.
        The default value is taken from the LensKit Library.
        """""
        reg = UniformFloatHyperparameter('reg', lower=0.001, upper=0.1, default_value=0.015)
        damping = UniformFloatHyperparameter('damping', lower=0.01, upper=1000, default_value=5, log=True)

        cs = ConfigurationSpace()
        cs.add([features, lrate, reg, damping])

        return cs



from lenskit.sklearn.svd import BiasedSVDScorer
from ConfigSpace import ConfigurationSpace, UniformIntegerHyperparameter, UniformFloatHyperparameter


class BiasedSVD(BiasedSVDScorer):
    def __init__(self, features, feedback="explicit", **kwargs):
        super().__init__(features=features, **kwargs)
        self.feedback = feedback
        self.features = features  # Store features as an instance variable for testing

    @staticmethod
    def get_default_configspace(number_item: int, **kwargs):
        """
               return default configurationspace
        """

        """
        The algorithm is based on the scikit TruncatedSVD algorithm and the hyperparameters are
        choosen based around the default values of LensKit. No further explaination could be distracted
        from scikit or the original paper.
        """
        n_items = number_item
        if n_items < 10000:
            features = UniformIntegerHyperparameter('features', lower=2, upper=n_items, default_value=n_items-1, log=True)
        else:
            # features = Integer('features', bounds=(2, 10000), default=1000, log=True)  # No default values given
            features = UniformIntegerHyperparameter('features', lower=2, upper=10000, default_value=1000, log=True)
        damping = UniformFloatHyperparameter('damping', lower=0.0001, upper=1000, default_value=5, log=True)

        cs = ConfigurationSpace()
        cs.add([features, damping])

        return cs



import numpy as np

from ConfigSpace import ConfigurationSpace

from lkauto.utils.get_model_from_cs import get_model_from_cs
from lkauto.optimization_strategies.bayesian_optimization import bayesian_optimization
from lkauto.optimization_strategies.random_search import random_search
from lkauto.utils.filer import Filer
from lkauto.ensemble.ensemble_builder import build_ensemble
from lkauto.preprocessing.preprocessing import preprocess_data
from lkauto.utils.logging import get_logger

# from lenskit.metrics.predict import rmse
# from lenskit.metrics.topn import ndcg
# from lenskit.algorithms import Predictor
# from lenskit import Recommender

from lenskit.metrics.predict import RMSE
from lenskit.metrics import NDCG
from lenskit.pipeline import Component
from lenskit.data import Dataset, ItemListCollection

from typing import Tuple


def get_best_prediction_model(train: Dataset,
                              validation: ItemListCollection = None,
                              cs: ConfigurationSpace = None,
                              optimization_metric=RMSE,
                              optimization_strategie: str = 'bayesian',
                              time_limit_in_sec: int = 2700,
                              num_evaluations: int = 500,
                              random_state=None,
                              split_folds: int = 1,
                              split_frac: float = 0.25,
                              split_strategie: str = 'user_based',
                              ensemble_size: int = 50,
                              minimize_error_metric_val: bool = True,
                              min_number_of_ratings: int = None,
                              max_number_of_ratings: int = None,
                              drop_duplicates: bool = False,
                              drop_na_values: bool = False,
                              user_column: str = 'user',
                              item_column: str = 'item',
                              rating_column: str = 'rating',
                              timestamp_col: str = 'timestamp',
                              include_timestamp: bool = True,
                              log_level: str = 'INFO',
                              filer: Filer = None) -> Tuple[Component, dict]:
    """
        returns the best Predictor found in the defined search time

         the find_best_explicit_configuration method will search the ConfigurationSpace
         for the best Predictor model configuration.
         Depending on the ConfigurationSpace parameter provided by the developer,
         performs three different use-cases.
         1. combined algorithm selection and hyperparameter configuration
         2. combined algorthm selection and hyperparameter configuration for a specific subset
            of algorithms and/or different parameter ranges
         3. hyperparameter selection for a specific algorithm.
         The hyperparameter and/or model selection process will be stopped after the time_limit_in_sec or (if set) after
         n_trials. The first one to be reached will stop the optimization.

        Parameters
        ----------
        train : pd.DataFrame
            Pandas Dataframe train split.
        validation : pd.DataFrame
            Pandas Dataframe validation split.
            if a validation split is provided, split_folds, split_strategy and split_frac will be ignored.
        cs : ConfigurationSpace
            ConfigurationSpace with all algorithms and parameter ranges defined.
        optimization_metric : function
            LensKit prediction accuracy metric to optimize for (either rmse or mae)
        optimization_strategie: str
            optimization strategie to use. Either bayesian or random_search
        time_limit_in_sec : int
            optimization search time limit in sec.
        num_evaluations : int
                number of samples to be used for optimization_strategy. Value can not be smaller than 6
                if no initial configuration is provided.
        random_state
            The random number generator or seed (see :py:func:`lenskit.util.rng`).
        split_folds : int
            number of folds of the inner split
        split_frac : float
            fraction of the inner split. If split_folds is to a value > 2, split_frac will be ignored.
            Value must be between 0 and 1.
        split_strategie : str
            split strategie to use. Either 'user_based' or 'item_based'
        ensemble_size : int
            number of models to be used for ensemble building.
        minimize_error_metric_val : bool
            if True, the optimization will try to minimize the error metric value.
            If False, the optimization will try to maximize the error metric value.
        min_number_of_ratings : int
            minimum number of ratings a user must have to be considered in the train dataset.
        max_number_of_ratings : int
            maximum number of ratings a user can have to be considered in the train dataset.
        drop_duplicates : bool
            if True, all duplicate rows will be dropped from the train dataset.
        drop_na_values : bool
            if True, all rows with NaN values will be dropped from the train dataset.
        user_column : str
            name of the user column in the train dataset.
        item_column : str
            name of the item column in the train dataset.
        rating_column : str
            name of the rating column in the train dataset.
        timestamp_col: str
            Name of the timestamp column
        include_timestamp: bool = True
            If True, the timestamp column will be included in the dataset
        log_level : str
            log level to use.
        filer : Filer
            filer to manage LensKit-Auto output

        Returns
        -------
        model : Predictor
            the best suited (untrained) predictor for the train dataset, cs parameters.
        incumbent : dict
            a dictionary containing the algorithm name and hyperparameter configuration of the returned model
   """

    logger = get_logger(level=log_level)

    # log parameters
    logger.info('---Starting LensKit-Auto---')
    if time_limit_in_sec is not None:
        logger.info('\t optimization_time: \t\t {} seconds'.format(time_limit_in_sec))
    if num_evaluations is not None:
        logger.info('\t num_evaluations: \t\t\t {}'.format(num_evaluations))
    logger.info('\t optimization_metric: \t\t {}'.format(optimization_metric.__name__))
    logger.info('\t optimization_strategie: \t {}'.format(optimization_strategie))

    # set num_evaluations to infinity if none is provided
    if num_evaluations is None:
        logger.debug('setting up num_evaluations to infinity')
        num_evaluations = np.inf

    # initialize filer if none is provided
    if filer is None:
        logger.debug('initializing filer')
        filer = Filer()

    # set random_state if none is provided
    if random_state is None:
        logger.debug('initializing random_state')
        random_state = 42

    # set split_folds to 1 if validation is not None
    if validation is not None:
        split_folds = 1

    # preprocess data
    preprocess_data(data=train,
                    user_col=user_column,
                    item_col=item_column,
                    rating_col=rating_column,
                    timestamp_col=timestamp_col,
                    include_timestamp=include_timestamp,
                    min_interactions_per_user=min_number_of_ratings,
                    max_interactions_per_user=max_number_of_ratings,
                    drop_na_values=drop_na_values,
                    drop_duplicates=drop_duplicates)

    # decide which optimization strategy to use
    if optimization_strategie == 'bayesian':
        incumbent, top_n_runs = bayesian_optimization(train=train,
                                                      cs=cs,
                                                      user_feedback='explicit',
                                                      validation=validation,
                                                      optimization_metric=optimization_metric,
                                                      time_limit_in_sec=time_limit_in_sec,
                                                      num_evaluations=num_evaluations,
                                                      random_state=random_state,
                                                      split_folds=split_folds,
                                                      split_frac=split_frac,
                                                      split_strategie=split_strategie,
                                                      ensemble_size=ensemble_size,
                                                      minimize_error_metric_val=minimize_error_metric_val,
                                                      filer=filer)
    elif optimization_strategie == 'random_search':
        incumbent, top_n_runs = random_search(train=train,
                                              cs=cs,
                                              user_feedback='explicit',
                                              validation=validation,
                                              optimization_metric=optimization_metric,
                                              time_limit_in_sec=time_limit_in_sec,
                                              num_evaluations=num_evaluations,
                                              random_state=random_state,
                                              split_folds=split_folds,
                                              split_frac=split_frac,
                                              split_strategie=split_strategie,
                                              ensemble_size=ensemble_size,
                                              minimize_error_metric_val=minimize_error_metric_val,
                                              filer=filer)
    else:
        raise ValueError('optimization_strategie must be either bayesian or random_search')

    # save top_n_runs to csv
    filer.save_dataframe_as_csv(top_n_runs, '', 'top_n_runs')

    logger.info('--Start Postrprocessing--')
    if ensemble_size > 1:
        model, incumbent = build_ensemble(train=train, top_n_runs=top_n_runs,
                                          filer=filer,
                                          ensemble_size=ensemble_size,
                                          lenskit_metric=optimization_metric,
                                          maximize_metric=(not minimize_error_metric_val))
        logger.info('--Best Model--')
        logger.info('GES Ensemble Model')
    else:
        # build model from best model configuration found by SMAC
        model = get_model_from_cs(incumbent, feedback='explicit')
        incumbent = incumbent.get_dictionary()
        logger.info('--Best Model--')
        logger.info(incumbent)

    logger.info('---LensKit-Auto finished---')

    # return model and model configuration
    return model, incumbent


def get_best_recommender_model(train: Dataset,
                               validation: ItemListCollection = None,
                               cs: ConfigurationSpace = None,
                               optimization_metric=NDCG,
                               optimization_strategie: str = 'bayesian',
                               time_limit_in_sec: int = 60,
                               num_evaluations: int = 500,
                               random_state=None,
                               split_folds: int = 1,
                               split_frac: float = 0.25,
                               split_strategie: str = 'user_based',
                               minimize_error_metric_val: bool = False,
                               num_recommendations: int = 10,
                               min_interactions_per_user: int = None,
                               max_interactions_per_user: int = None,
                               drop_duplicates: bool = False,
                               drop_na_values: bool = False,
                               user_column: str = 'user',
                               item_column: str = 'item',
                               rating_column: str = 'rating',
                               timestamp_col: str = 'timestamp',
                               include_timestamp: bool = True,
                               log_level: str = 'INFO',
                               filer: Filer = None) -> Tuple[Component, dict]:
    """
        returns the best Recommender found in the defined search time

         the find_best_implicit_configuration method will search the ConfigurationSpace
         for the best Recommender model configuration.
         Depending on the ConfigurationSpace parameter provided by the developer,
         performs three different use-cases.
         1. combined algorithm selection and hyperparameter configuration
         2. combined algorthm selection and hyperparameter configuration for a specific subset
            of algorithms and/or different parameter ranges
         3. hyperparameter selection for a specific algorithm.
         The hyperparameter and/or model selection process will be stopped after the time_limit_in_sec or (if set) after
         n_trials. The first one to be reached will stop the optimization.

        Parameters
        ----------
        train : pd.DataFrame
            Pandas Dataframe train split.
        validation : pd.DataFrame
            Pandas Dataframe validation split.
            if a validation split is provided, split_folds, split_strategy and split_frac will be ignored.
        cs : ConfigurationSpace
            ConfigurationSpace with all algorithms and parameter ranges defined.
        optimization_strategie: str
            optimization strategie to use. Either bayesian or random_search
        optimization_metric : function
            LensKit recommender metric to optimize for
        time_limit_in_sec : int
            search time limit.
        num_evaluations : int
                number of samples to be used for optimization_strategy. Value can not be smaller than 6
        random_state
            The random number generator or seed (see :py:func:`lenskit.util.rng`).
        split_folds : int
            number of folds of the inner split
        split_frac : float
            fraction of the inner split. If split_folds is not None, split_frac will be ignored.
            Value must be between 0 and 1.
        split_strategie : str
            split strategie to use. Either 'user_based' or 'item_based'.
        minimize_error_metric_val : bool
            if True, the optimization_metric value will be minimized.
            If False, the optimization_metric value will be maximized.
        num_recommendations : int
            number of recommendations to be evaluted per user. Value must be greater than 0.
        min_interactions_per_user : int
            minimum number of ratings a user must have to be considered in the train dataset.
        max_interactions_per_user : int
            maximum number of ratings a user can have to be considered in the train dataset.
        drop_duplicates : bool
            if True, all duplicate rows will be dropped from the train dataset.
        drop_na_values : bool
            if True, all rows with NaN values will be dropped from the train dataset.
        user_column : str
            name of the user column in the train dataset.
        item_column : str
            name of the item column in the train dataset.
        rating_column : str
            name of the rating column in the train dataset.
        timestamp_col: str
            Name of the timestamp column
        include_timestamp: bool = True
            If True, the timestamp column will be included in the dataset
        log_level : str
            log level to use.
        filer : Filer
            filer to manage LensKit-Auto output

        Returns
        -------
        model : Predictor
            the best suited (untrained) predictor for the train dataset, cs parameters.
        incumbent : dict
            a dictionary containing the algorithm name and hyperparameter configuration of the returned model
    """

    logger = get_logger(level=log_level)

    # log parameters
    logger.info('---Starting LensKit-Auto---')
    if time_limit_in_sec is not None:
        logger.info('\t optimization_time: \t\t {} seconds'.format(time_limit_in_sec))
    if num_evaluations is not None:
        logger.info('\t num_evaluations: \t\t\t {}'.format(num_evaluations))
    logger.info('\t optimization_metric: \t\t {}@{}'.format(optimization_metric.__name__, num_recommendations))
    logger.info('\t optimization_strategie: \t {}'.format(optimization_strategie))

    # set num_evaluations to infinity if none is provided
    if num_evaluations is None:
        logger.debug('num_evaluations is None. Setting num_evaluations to infinity.')
        num_evaluations = np.inf

    # initialize filer if none is provided
    if filer is None:
        logger.debug('filer is None. Initializing filer.')
        filer = Filer()

    # set random state if none is provided
    if random_state is None:
        logger.debug('random_state is None. Initializing random_state.')
        random_state = 42

    # set split_folds to 1 if validation is not None
    if validation is not None:
        split_folds = 1

    # preprocess data
    preprocess_data(data=train,
                    user_col=user_column,
                    item_col=item_column,
                    rating_col=rating_column,
                    timestamp_col=timestamp_col,
                    include_timestamp=include_timestamp,
                    min_interactions_per_user=min_interactions_per_user,
                    max_interactions_per_user=max_interactions_per_user,
                    drop_na_values=drop_na_values,
                    drop_duplicates=drop_duplicates)

    # define optimization strategie to use
    if optimization_strategie == 'bayesian':
        incumbent = bayesian_optimization(train=train,
                                          validation=validation,
                                          cs=cs,
                                          user_feedback='implicit',
                                          optimization_metric=optimization_metric,
                                          time_limit_in_sec=time_limit_in_sec,
                                          num_evaluations=num_evaluations,
                                          random_state=random_state,
                                          split_folds=split_folds,
                                          split_frac=split_frac,
                                          split_strategie=split_strategie,
                                          minimize_error_metric_val=minimize_error_metric_val,
                                          num_recommendations=num_recommendations,
                                          filer=filer)
    elif optimization_strategie == 'random_search':
        incumbent = random_search(train=train,
                                  validation=validation,
                                  cs=cs,
                                  user_feedback='implicit',
                                  optimization_metric=optimization_metric,
                                  time_limit_in_sec=time_limit_in_sec,
                                  num_evaluations=num_evaluations,
                                  random_state=random_state,
                                  split_folds=split_folds,
                                  split_frac=split_frac,
                                  split_strategie=split_strategie,
                                  minimize_error_metric_val=minimize_error_metric_val,
                                  num_recommendations=num_recommendations,
                                  filer=filer)
    else:
        raise ValueError('optimization_strategie must be either bayesian or random_search')

    logger.info('--Start Postrprocessing--')

    # build model from best model configuration found by SMAC
    model = get_model_from_cs(incumbent, feedback='implicit')
    incumbent = incumbent.get_dictionary()

    logger.info('--Best Model--')
    logger.info(incumbent)

    logger.info('---LensKit-Auto finished---')

    # return model and model configuration
    return model, incumbent



import unittest

import ConfigSpace as CS

from lkauto.algorithms.funksvd import FunkSVD


class TestFunkSVD(unittest.TestCase):
    def test_init_givenObjectInitialized_ObjectInitializedCorrectlyExpected(self):
        features = 100
        funk_svd = FunkSVD(features)
        self.assertIsInstance(funk_svd, FunkSVD)
        self.assertEqual(features, funk_svd.features)

    def test_defaultConfigspace_GivenFunctionCalled_correctConfigSpaceReturnedExpected(self):
        cs = FunkSVD.get_default_configspace()
        params = list(cs.values())
        self.assertIsInstance(cs, CS.ConfigurationSpace)
        self.assertTrue(any(
            param.name == "features" and param.default_value == 1000 and param.lower == 2 and param.upper == 10000 for
            param in params))
        self.assertTrue(any(
            param.name == "lrate" and param.default_value == 0.001 and param.lower == 0.0001 and param.upper == 0.01 for
            param in params))
        self.assertTrue(any(
            param.name == "reg" and param.default_value == 0.015 and param.lower == 0.001 and param.upper == 0.1 for
            param in params))
        self.assertTrue(any(
            param.name == "damping" and param.default_value == 5 and param.lower == 0.01 and param.upper == 1000 for
            param in params))


if __name__ == '__main__':
    unittest.main()



from typing import Union

from ConfigSpace import ConfigurationSpace

from lenskit.als import BiasedMFScorer
from lenskit.als import ImplicitMFScorer
from lenskit.basic import BiasScorer
from lenskit.funksvd import FunkSVDScorer
from lenskit.knn import ItemKNNScorer
from lenskit.sklearn.svd import BiasedSVDScorer
from lenskit.knn import UserKNNScorer
# from lenskit.scored import Scorer


def get_model_from_cs(cs: ConfigurationSpace,
                      feedback: str,
                      # fallback_model=Bias(),
                      random_state: int = 42) -> Union[
                          ItemKNNScorer,
                          UserKNNScorer,
                          FunkSVDScorer,
                          BiasedSVDScorer,
                          BiasedMFScorer,
                          BiasScorer,
                          ImplicitMFScorer]:
    """ builds a Predictor model defined in ConfigurationSpace

        Parameters
        ----------
        cs : ConfigurationSpace
            configuration space containing information to build a model
        feedback : str
            feedback type, either 'explicit' or 'implicit'
        random_state: int
            random state to use

        Returns
        ----------
        model: Union[ItemKNNScorer, UserKNNScorer, FunkSVDScorder, BiasedSVDScorer, BiasedMFScorer, BiasScorer, ImplicitMFScorer]
            Model build with the config_space information
    """

    # check if feedback value is valid
    if (feedback != 'explicit') and (feedback != 'implicit'):
        raise ValueError("Unknown feedback type: {}".format(feedback))

    # get algorithm name
    algo_name = cs.get('algo')
    config = {key.replace("{}:".format(algo_name), ""): value for key, value in cs.items()}
    del config['algo']

    # ItemItem
    if algo_name == 'ItemItem':
        # model = ItemKNNScorer(feedback=feedback, **config)
        model = ItemKNNScorer(feedback=feedback)
    # UserUser
    elif algo_name == 'UserUser':
        # model = UserKNNScorer(feedback=feedback, **config)
        model = UserKNNScorer(feedback=feedback)
    # FunkSVD
    elif algo_name == 'FunkSVD':
        # model = FunkSVDScorer(random_state=random_state, **config)
        model = FunkSVDScorer(feedback=feedback)
    # BiasedSVD
    elif algo_name == 'BiasedSVD':
        # model = BiasedSVDScorer(**config)
        model = BiasedSVDScorer(feedback=feedback)
    # ALSBiasedMF
    elif algo_name == 'ALSBiasedMF':
        # reg_touple = (float(config.pop('ureg')), float(config.pop('ireg')))
        # del config['ureg']
        # del config['ireg']
        # model = BiasedMFScorer(reg=reg_touple, rng_spec=random_state, **config)
        model = BiasedMFScorer(feedback=feedback)
    # Biased
    elif algo_name == 'Bias':
        # damping_touple = (config.pop('user_damping'), config.pop('item_damping'))
        # del config['user_damping']
        # del config['item_damping']
        # model = BiasScorer(damping=damping_touple, **config)
        model = BiasScorer()
    # ImplicitMF
    elif algo_name == 'ImplicitMF':
        # reg_touple = (float(config.pop('ureg')), float(config.pop('ireg')))
        # del config['ureg']
        # del config['ireg']
        # model = ImplicitMFScorer(reg=reg_touple, rng_spec=random_state, **config)
        model = ImplicitMFScorer(feedback=feedback)
    else:
        raise ValueError("Unknown algorithm: {}".format(algo_name))

    final_model = model
    return final_model



from typing import Iterator

from lenskit.splitting import crossfold_records, crossfold_users, sample_records, SampleFrac, TTSplit
from lenskit.data import Dataset


def validation_split(data: Dataset, strategy: str = 'user_based', num_folds: int = 1,
                     frac: float = 0.25, random_state=42) -> Iterator[TTSplit]:
    """
    Returns the Train-Test-Split for the given Dataset

    Parameters
    ----------
    data : Dataset
        Lenskit Dataset with the data to be split.
    strategy : str
        cross validation strategy (either "user_based" or "row_based")
    num_folds : int
        number of folds for the validation split cross validation
    frac : float
        fraction of the dataset to be used for the validation split. If num_folds > 1, the fraction value
        will be ignored.
    random_state : int
        random state for the validation split

    Returns
    -------
    Iterator[TTSplit]
        The Train-Test-Split for the given Dataset
    """
    # decide which validation split strategy to use
    if strategy == 'user_based':
        return user_based_validation_split(data=data, num_folds=num_folds, frac=frac, random_state=random_state)
    elif strategy == 'row_based':
        return row_based_validation_split(data=data, num_folds=num_folds, frac=frac, random_state=random_state)
    else:
        raise ValueError(f"Unknown validation split strategy: {strategy}")


def row_based_validation_split(data: Dataset, num_folds: int = 1, frac: float = 0.25, random_state=42) -> Iterator[TTSplit]:
    """
    Returns a Train-Test-Split for the given data.

    Parameters
    ----------
    data : Dataset
        Lenskit Dataset with the data to be split.
    num_folds : int
        number of folds for the validation split cross validation
    frac : float
        fraction of the dataset to be used for the validation split. If num_folds > 1, the fraction value
        will be ignored.
    random_state : int
        random state for the validation split

    Returns
    -------
    Iterator[TTSPlit]
        Train-Tet-Split for the given data.
    """

    if num_folds < 2:
        return __holdout_validation_split(data=data, frac=frac, random_state=random_state)
    else:
        return __row_based_k_fold_validation_split(data=data, num_folds=num_folds, random_state=random_state)


def user_based_validation_split(data: Dataset, num_folds: int = 1, frac: float = 0.25, random_state=42) -> Iterator[TTSplit]:
    """
    Parameters
    ----------
    data : Dataset
        Lenskit Dataset with the data to be split.
    num_folds : int
        number of folds for the validation split cross validation
    frac : float
        fraction of the dataset to be used for the validation split. If num_folds > 1, the fraction value
        will be ignored.
    random_state : int
        random state for the validation split

    Returns
    -------
    Iterator[TTSplit]
        Train-Test-Split for the given data.
    """

    if num_folds < 2:
        return __holdout_validation_split(data=data, frac=frac, random_state=random_state)
    else:
        return __user_based_crossfold_validation_split(data=data, num_folds=num_folds)


def __holdout_validation_split(data: Dataset, frac: float, random_state=42):
    """
    Returns a Train-Test-Split for the given data.

    Parameters
    ----------
    data : Dataset
        Lenskit Dataset with the data to be split.
    frac : float
        fraction of the dataset to be used for the validation split. If num_folds > 1, the fraction value
        will be ignored.
    random_state : int
        random state for the validation split

    Returns
    -------
    Iterator[TTSplit]
        Train-Test-Split for the given data. Should only contain one fold.
    """

    splits = sample_records(data=data, size=int(data.interaction_count * frac), rng=random_state)

    if hasattr(splits, "_iter__"):
        return splits
    else:
        return iter([splits])


def __row_based_k_fold_validation_split(data: Dataset, num_folds: int, random_state):
    """
    Returns a Train-Test-Split for the given data.

    Parameters
    ----------
    data : Dataset
        Lenskit Dataset with the data to be split.
    """

    splits = crossfold_records(data=data, partitions=num_folds, rng=random_state)
    return splits


def __user_based_crossfold_validation_split(data: Dataset, num_folds) -> Iterator[TTSplit]:
    """
    Returns a Train-Test-Split for the given data.

    Parameters
    ----------
    data : Dataset
        Pandas Dataframe with the data to be split.
    num_folds : int
        number of folds for the validation split cross validation

    Returns
    -------
    Iterator[TTSplit]
        Train-Test-Split for the given data.
    """

    return crossfold_users(data=data, partitions=num_folds, method=SampleFrac(0.2))



from typing import Iterator, Union
from ConfigSpace import Categorical
from ConfigSpace import ConfigurationSpace
from lenskit.data import Dataset
from lenskit.splitting import TTSplit

from lkauto.algorithms.als import BiasedMF
from lkauto.algorithms.als import ImplicitMF
from lkauto.algorithms.bias import Bias
from lkauto.algorithms.funksvd import FunkSVD
from lkauto.algorithms.item_knn import ItemItem
from lkauto.algorithms.svd import BiasedSVD
from lkauto.algorithms.user_knn import UserUser


def get_default_configuration_space(data: Union[Dataset, Iterator[TTSplit]],
                                    feedback: str,
                                    val_fold_indices=None,
                                    validation: Iterator[TTSplit] = None,
                                    random_state=42) -> ConfigurationSpace:
    """
        returns the default configuration space for all included rating prediction algorithms

        Parameters
        ----------
        data: Union[Dataset, Iterator[TTSplit]]
            data to use (either a Dataset or a Train-Test-Pair)
        val_fold_indices
            validation fold indices
        validation: Iterator[TTSplit]
            validation data (provided by user)
        feedback : str
            feedback type (either 'explicit' or 'implicit')
        random_state: int
            random state to use

        Returns
        ------
        ConfigurationSpace
            The default configuration space
    """

    if feedback == 'explicit':
        algo_list = ['ItemItem', 'UserUser', 'FunkSVD', 'BiasedSVD', 'ALSBiasedMF', 'Bias']
    elif feedback == 'implicit':
        algo_list = ['ItemItem', 'UserUser', 'ImplicitMF',]
    else:
        raise ValueError("Unknown feedback type: {}".format(feedback))

    # get minimum number of items and users for the given train split
    num_items = 0
    num_users = 0

    if validation is None and not isinstance(data, Dataset):
        for fold in data:
            if fold.train.item_count < num_items or num_items == 0:
                num_items = fold.train.item_count
            if fold.train.user_count < num_users or num_users == 0:
                num_users = fold.train.user_count
    else:
        if data.item_count < num_items or num_items == 0:
            num_items = data.item_count
        if data.user_count < num_users or num_users == 0:
            num_users = data.user_count

    # define configuration space
    cs = ConfigurationSpace(
        seed=random_state,
        space={
            "algo": Categorical("algo", algo_list, default="ItemItem"),
        }
    )

    for algo in algo_list:
        if algo == 'UserUser':
            model = UserUser
        elif algo == 'ItemItem':
            model = ItemItem
        elif algo == 'FunkSVD':
            model = FunkSVD
        elif algo == 'BiasedSVD':
            model = BiasedSVD
        elif algo == 'ALSBiasedMF':
            model = BiasedMF
        elif algo == 'Bias':
            model = Bias
        elif algo == 'ImplicitMF':
            model = ImplicitMF
        else:
            raise ValueError("Unknown algorithm: {}".format(algo))

        # add configuration space of algorithm
        cs.add_configuration_space(
            prefix=algo,
            delimiter=":",
            configuration_space=model.get_default_configspace(number_user=num_users, number_item=num_items),
            parent_hyperparameter={"parent": cs["algo"], "value": algo},
        )

    return cs



from lenskit.basic.bias import BiasScorer
from ConfigSpace import ConfigurationSpace, UniformFloatHyperparameter


class Bias(BiasScorer):
    def __init__(self, feedback="explicit", **kwargs):
        super().__init__(**kwargs)

        self.feedback = feedback

    @staticmethod
    def get_default_configspace(number_item: int, number_user: int):
        """
            return default configurationspace
        """

        """
        LensKit does not give any hint on how to setup the damping values for the Bias algorithm.
        Therefore we evaluated the algorithms performance on 70 datasets to come up with the following ranges
        """
        # lower bound 0 does not work because of log=True
        item_damping = UniformFloatHyperparameter('item_damping', lower=1e-5*number_item, upper=85*number_item,
                                                  default_value=0.0025*number_item, log=True)
        user_damping = UniformFloatHyperparameter('user_damping', lower=1e-5*number_user, upper=85*number_user,
                                                  default_value=0.0025*number_user, log=True)

        cs = ConfigurationSpace()
        cs.add([item_damping, user_damping])

        return cs



import numpy as np
import pandas as pd
from ConfigSpace import ConfigurationSpace

from lenskit.batch import predict
from lenskit.data import Dataset
from lenskit.metrics import RunAnalysis
import logging

from lenskit.pipeline import predict_pipeline

from lkauto.utils.filer import Filer
from lkauto.utils.get_model_from_cs import get_model_from_cs
from lkauto.utils.validation_split import validation_split


class ImplicitEvaler:
    """ImplicitEvaler

            the ImplicitEvaler class handles the evaluation of the optimization tool.
            An Evaluation run consists of training a model and predict the performance on a validation split.

            Attributes
            ----------
            train : Dataset
                lenskit Dataset containing the train split.
            optimization_metric: function
                LensKit top-n metric used to evaluate the model
            filer : Filer
                Filer to organize the output.
            validation : ItemListCollection
                ItemListCollection containing the validation split.
             random_state :
                 The random number generator or seed (see :py:func:`lenskit.util.rng`).
            split_folds :
                The number of folds of the validation split
            split_strategy :
                The strategy used to split the data. Possible values are 'user_based' and 'row_based'
            split_frac :
                The fraction of the data used for the validation split. If the split_folds value is greater than 1,
                this value is ignored.
            num_recommendations :
                The number of recommendations to be made and evaluated for each user.
            minimize_error_metric_val :
                If True, the error metric is minimized. If False, the error metric is maximized. This parameter needs to be
                set in corelation with the optimization metric.

             Methods
            ----------
            evaluate_explicit(config_space: ConfigurationSpace) -> float
    """

    def __init__(self,
                 train: Dataset,
                 optimization_metric,
                 filer: Filer,
                 validation=None,
                 random_state=42,
                 split_folds: int = 1,
                 split_strategy: str = 'user_based',
                 split_frac: float = 0.25,
                 num_recommendations: int = 10,
                 minimize_error_metric_val: bool = True,
                 ) -> None:
        self.logger = logging.getLogger('lenskit-auto')
        self.train = train
        self.validation = validation
        self.optimization_metric = optimization_metric
        self.random_state = random_state
        self.split_folds = split_folds
        self.split_strategy = split_strategy
        self.split_frac = split_frac
        self.filer = filer
        self.num_recommendations = num_recommendations
        self.minimize_error_metric_val = minimize_error_metric_val
        self.run_id = 0
        # create validation split
        if self.validation is None:
            self.val_fold_indices = validation_split(data=self.train,
                                                     strategy=self.split_strategy,
                                                     num_folds=self.split_folds,
                                                     frac=self.split_frac,
                                                     random_state=self.random_state)
        else:
            self.val_fold_indices = None

    def evaluate(self, config_space: ConfigurationSpace) -> float:
        """ evaluates model defined in config_space

            The config_space parameter defines a model.
            This model is build, trained and evaluated with the validation split.

            Parameters
            ----------
            config_space : ConfigurationSpace
                configuration space containing information to build a model

            Returns
            ----------
            validation_error : float
                the error of the considered model
        """
        output_path = 'eval_runs/'
        self.run_id += 1
        metric_scores = np.array([])
        validation_data = pd.DataFrame()
        scores = None

        # get model form configuration space
        model = get_model_from_cs(config_space, feedback='implicit')

        if self.validation is None:
            for fold in self.val_fold_indices:
                validation_train = fold.train
                validation_test = fold.test

                pipeline = predict_pipeline(scorer=model)
                fit_pipeline = pipeline.clone()
                fit_pipeline.train(validation_train)

                recs = predict(fit_pipeline, validation_test)

                # create rec list analysis
                rla = RunAnalysis()
                rla.add_metric(self.optimization_metric)

                # compute scores
                scores = rla.measure(recs, validation_test)

                # store data
                validation_data = pd.concat([validation_data, recs.to_df()], axis=0)
                # the first (index 0) column should contain the means for the metrics (rows)
                metric_scores = np.append(metric_scores, scores.list_summary().loc[self.optimization_metric.__name__, "mean"])
        else:
            for fold in range(self.split_folds):
                validation_train = self.train
                validation_test = self.validation

                pipeline = predict_pipeline(scorer=model)
                fit_pipeline = pipeline.clone()
                fit_pipeline.train(validation_train)

                recs = predict(fit_pipeline, validation_test)

                # create rec list analysis
                rla = RunAnalysis()
                rla.add_metric(self.optimization_metric)

                # compute scores
                scores = rla.measure(recs, validation_test)

                # store data
                validation_data = pd.concat([validation_data, recs.to_df()], axis=0)
                metric_scores = np.append(metric_scores, scores.list_summary().loc[self.optimization_metric.__name__, "mean"])

        # save validation data
        self.filer.save_validataion_data(config_space=config_space,
                                         predictions=validation_data,
                                         metric_scores=metric_scores,
                                         output_path=output_path,
                                         run_id=self.run_id)

        # store score mean and subtract by 1 to enable SMAC to minimize returned value
        # the first (index 0) column should contain the means for the metrics (rows)
        validation_error = scores.list_summary().loc[self.optimization_metric.__name__, "mean"] - 1

        self.logger.info('Run ID: ' + str(self.run_id) + ' | ' + str(config_space.get('algo')) + ' | ' +
                         self.optimization_metric.__name__ + '@{}'.format(self.num_recommendations) + ': '
                         + str(validation_error))
        self.logger.debug(str(config_space))

        if self.minimize_error_metric_val:
            return validation_error
        else:
            return 1 - validation_error



import unittest

import ConfigSpace as CS

from lkauto.algorithms.bias import Bias


class TestBias(unittest.TestCase):
    def test_init_givenObjectInitialized_ObjectInitializedCorrectlyExpected(self):
        bias = Bias()
        self.assertIsInstance(bias, Bias)

    def test_defaultConfigspace_GivenFunctionCalled_correctConfigSpaceReturnedExpected(self):
        number_item = 10
        number_user = 100
        cs = Bias.get_default_configspace(number_item=number_item, number_user=number_user)
        params = list(cs.values())
        self.assertIsInstance(cs, CS.ConfigurationSpace)
        self.assertTrue(any(
            param.name == "item_damping" and param.default_value == 0.025 and param.lower == 1e-4 and param.upper == 850
            for
            param in params))
        self.assertTrue(any(
            param.name == "user_damping" and param.default_value == 0.25 and param.lower == 1e-3 and param.upper == 8500
            for
            param in params))


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd
import time

from ConfigSpace import ConfigurationSpace, Configuration

from lenskit.data import Dataset, ItemListCollection

from lkauto.explicit.explicit_evaler import ExplicitEvaler
from lkauto.implicit.implicit_evaler import ImplicitEvaler
from lkauto.utils.get_default_configurations import get_default_configurations
from lkauto.utils.filer import Filer
from lkauto.utils.get_default_configuration_space import get_default_configuration_space

from typing import Tuple
import logging


def random_search(cs: ConfigurationSpace,
                  train: Dataset,
                  user_feedback: str,
                  optimization_metric,
                  filer: Filer,
                  validation: ItemListCollection = None,
                  time_limit_in_sec: int = 3600,
                  num_evaluations: int = None,
                  split_folds: int = 1,
                  split_strategie: str = 'user_based',
                  split_frac: float = 0.25,
                  ensemble_size: int = 50,
                  minimize_error_metric_val: bool = True,
                  num_recommendations: int = 10,
                  random_state=42) -> Tuple[Configuration, pd.DataFrame]:
    """
        returns the best configuration found by random search

         The random_search method will randomly search through the ConfigurationSpace to find the
         best performing configuration for the given train split. The ConfigurationSpace can consist of
         hyperparameters for a single algorithm or a combination of algorithms.

        Parameters
        ----------
        train : pd.DataFrame
            Pandas Dataframe train split.
        cs : ConfigurationSpace
            ConfigurationSpace with all algorithms and hyperparameter ranges defined.
        num_evaluations: int
            number of samples to be randomly drawn from the ConfigurationSpace
        optimization_metric : function
            LensKit prediction accuracy metric to optimize for (either rmse or mae)
        minimize_error_metric_val : bool
            Bool that decides if the error metric should be minimized or maximized.
        user_feedback : str
            Defines if the dataset contains explicit or implicit feedback.
        random_state: int
        filer : Filer
            filer to manage LensKit-Auto output
        validation : pd.DataFrame
            Pandas Dataframe validation split.
        time_limit_in_sec
            time limit in seconds for the optimization process
        split_folds : int
            number of folds for the validation split cross validation
        split_strategie : str
            cross validation strategie (user_based or row_based)
        split_frac : float
            fraction of the dataset to be used for the validation split. If num_folds > 1, the fraction value
            will be ignored.
        ensemble_size : int
            number of models to be used in the ensemble of rating prediction tasks. This value will be ignored
            for recommender tasks.
        num_recommendations : int
            number of recommendations to be made for each user. This value will be ignored for rating prediction.

        Returns
        -------
        best_configuration : Configuration
            the best suited (algorithm and) hyperparameter configuration for the train dataset.
        top_n_runs : pd.DataFrame
            pandas dataframe containing the top n runs of the random search.
   """
    logger = logging.getLogger('lenskit-auto')
    logger.info('--Start Random Search--')

    # initialize variables to keep track of the best configuration
    best_configuration = None
    if minimize_error_metric_val:
        best_error_score = 0
    else:
        best_error_score = np.inf

    # initialize evaler based on user feedback
    if user_feedback == "explicit":
        evaler = ExplicitEvaler(train=train,
                                optimization_metric=optimization_metric,
                                filer=filer,
                                validation=validation,
                                random_state=random_state,
                                split_folds=split_folds,
                                split_strategy=split_strategie,
                                split_frac=split_frac,
                                ensemble_size=ensemble_size,
                                minimize_error_metric_val=minimize_error_metric_val)
    elif user_feedback == 'implicit':
        evaler = ImplicitEvaler(train=train,
                                optimization_metric=optimization_metric,
                                filer=filer,
                                validation=validation,
                                random_state=random_state,
                                split_folds=split_folds,
                                split_strategy=split_strategie,
                                split_frac=split_frac,
                                num_recommendations=num_recommendations,
                                minimize_error_metric_val=minimize_error_metric_val)
    else:
        raise ValueError('feedback must be either explicit or implicit')

    # get pre-defined ConfiguraitonSpace if none is provided
    if cs is None:
        logger.debug('initializing default ConfigurationSpace')
        cs = get_default_configuration_space(data=train,
                                             val_fold_indices=evaler.train_test_splits,
                                             validation=validation,
                                             feedback='explicit',
                                             random_state=random_state)

    # get LensKit default configurations
    initial_configuraition = get_default_configurations(cs)

    # run for a specified number of iterations or until time limit is reached
    if num_evaluations == 0:
        # track time to support random_search on a time base
        start_time = time.time()

        # track configurations that have already been tested
        configuration_list = []

        # loop through random spampled configurations
        while time.time() - start_time < time_limit_in_sec:
            # initialize config
            config = None
            # check if all initial configurations have been tested
            if all(x in configuration_list for x in initial_configuraition):
                # random sample configuration from configuration space
                config = cs.sample_configuration()
            else:
                # pop configuration from initial configurations
                config = initial_configuraition.pop(0)

            # check if configuration has already been tested
            if config not in configuration_list:
                # calculate error for the configuration
                error = evaler.evaluate(config)

                # keep track of best performing configuration
                if minimize_error_metric_val:
                    if error > best_error_score:
                        best_error_score = error
                        best_configuration = config
                else:
                    if error < best_error_score:
                        best_error_score = error
                        best_configuration = config
                # add configuration to list of tested configurations
                configuration_list.append(config)
    else:
        configuration_set = set()

        # add initial configurations to set
        for config in initial_configuraition:
            if len(configuration_set) < num_evaluations:
                configuration_set.add(config)

        # add configurations to set until the set contains num_evaluations configurations
        while len(configuration_set) < num_evaluations:
            configuration_set.add(cs.sample_configuration())

        # track time to support random_search on a time base
        start_time = time.time()

        # loop through random spampled configurations
        for config in configuration_set:
            # calculate error for the configuration
            error = evaler.evaluate(config)

            # keep track of best performing configuration
            if minimize_error_metric_val:
                if error > best_error_score:
                    best_error_score = error
                    best_configuration = config
            else:
                if error < best_error_score:
                    best_error_score = error
                    best_configuration = config

            # keep track of time
            if time.time() - start_time > time_limit_in_sec:
                break

    logger.info('--End Random Search--')

    if user_feedback == "explicit":
        filer.save_dataframe_as_csv(evaler.top_n_runs, '', 'top_n_runs')
        return best_configuration, evaler.top_n_runs
    elif user_feedback == 'implicit':
        return best_configuration
    else:
        raise ValueError('feedback must be either explicit or implicit')



import logging


def get_logger(name: str = 'lenskit-auto', level: str = logging.INFO) -> logging.Logger:
    """
    Returns a logger with the given name and level.

    Parameters
    ----------
    name : str
        name of the logger
    level : int
        level of the logger

    Returns
    -------
    logging.Logger
        logger with the given name and level
    """
    # logging.basicConfig(
    #     level=level,
    #     format="%(asctime)s %(levelname)s %(message)s",
    #     datefmt="%Y-%m-%d %H:%M:%S",
    # )

    logger = logging.getLogger(name=name)
    logger.setLevel(level=level)

    fh = logging.StreamHandler()
    fh_formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    fh.setFormatter(fh_formatter)

    logger.addHandler(fh)

    return logger



from pathlib import Path
from setuptools import setup, find_packages

SHORT_DESCRIPTION = """LensKit-Auto is built as a wrapper around the Python LensKit recommender-system library. It automates algorithm selection and hyper parameter optimization an can build ensemble models based on the LensKit models."""

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="lenskit-auto",
    version="0.1.2",
    author="Tobias Vente",
    python_requires=">=3.12, <3.13",
    packages=find_packages(),
    install_requires=[
        "smac>=2.3.1",
        "lenskit>=2025.2.0",
        "numpy>=2.2.6",
        "pandas>=2.2.3",
        "scikit-learn>=1.6.1",
        "scipy>=1.15.3",
        "numba>=0.61.2",
        "typing-extensions>=4.13.2",
        "matplotlib>=3.10.3",
    ],
    extras_require={
        "test": ["pytest>=8.4.1", "pytest-cov>=6.2.1"],
        "doc": ["sphinx>=4.2", "sphinx-rtd-theme>=1.0.0", "nbsphinx>=0.8.9"],
    },
    entry_points={},
    description=SHORT_DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ISG-Siegen/lenskit-auto",
    project_urls={"Documentation": "https://lenskit-auto.readthedocs.io"},
)



import unittest

import ConfigSpace as CS

from lkauto.algorithms.als import ImplicitMF, BiasedMF


class TestImplicitMF(unittest.TestCase):
    def test_init_givenObjectInitialized_ObjectInitializedCorrectlyExpected(self):
        expected_features = 10
        implicit_mf = ImplicitMF(expected_features)
        self.assertIsInstance(implicit_mf, ImplicitMF)
        self.assertEqual(expected_features, implicit_mf.features)

    def test_defaultConfigspace_GivenFunctionCalled_correctConfigSpaceReturnedExpected(self):
        cs = ImplicitMF.get_default_configspace()
        params = list(cs.values())
        self.assertIsInstance(cs, CS.ConfigurationSpace)
        self.assertTrue(any(
            param.name == "features" and param.default_value == 1000 and param.lower == 5 and param.upper == 10000 for
            param in params))
        self.assertTrue(any(
            param.name == "ureg" and param.default_value == 0.1 and param.lower == 0.01 and param.upper == 0.1 for param
            in params))
        self.assertTrue(any(
            param.name == "ireg" and param.default_value == 0.1 and param.lower == 0.01 and param.upper == 0.1 for param
            in params))


class TestBiasedMF(unittest.TestCase):
    def test_init_givenObjectInitialized_ObjectInitializedCorrectlyExpected(self):
        expected_features = 10
        biased_mf = BiasedMF(expected_features)
        self.assertIsInstance(biased_mf, BiasedMF)
        self.assertEqual(expected_features, biased_mf.features)

    def test_defaultConfigspace_GivenFunctionCalled_correctConfigSpaceReturnedExpected(self):
        cs = BiasedMF.get_default_configspace()
        params = list(cs.values())
        self.assertIsInstance(cs, CS.ConfigurationSpace)
        self.assertTrue(any(
            param.name == "features" and param.default_value == 1000 and param.lower == 2 and param.upper == 10000 for
            param in params))
        self.assertTrue(any(
            param.name == "ureg" and param.default_value == 0.1 and param.lower == 0.01 and param.upper == 0.1 for param
            in params))
        self.assertTrue(any(
            param.name == "ireg" and param.default_value == 0.1 and param.lower == 0.01 and param.upper == 0.1 for param
            in params))
        self.assertTrue(any(
            param.name == "bias" and param.default_value for param in
            params))


if __name__ == '__main__':
    unittest.main()



import unittest

import ConfigSpace as CS

from lkauto.algorithms.user_knn import UserUser


class TestUserKNN(unittest.TestCase):
    def test_init_givenObjectInitialized_ObjectInitializedCorrectlyExpected(self):
        max_n = 100
        min_n = 5
        min_s = 0.001
        user_user = UserUser(max_n, min_nbrs=min_n, min_sim=min_s)
        self.assertIsInstance(user_user, UserUser)
        self.assertEqual(max_n, user_user.max_nbrs)
        self.assertEqual(min_n, user_user.min_nbrs)
        self.assertEqual(min_s, user_user.min_sim)

    def test_defaultConfigspace_GivenFunctionCalled_correctConfigSpaceReturnedExpected(self):
        cs = UserUser.get_default_configspace()
        params = list(cs.values())
        self.assertIsInstance(cs, CS.ConfigurationSpace)
        self.assertTrue(any(
            param.name == "max_nbrs" and param.default_value == 1000 and param.lower == 1 and param.upper == 10000 for
            param in params))
        self.assertTrue(any(
            param.name == "min_nbrs" and param.default_value == 1 and param.lower == 1 and param.upper == 1000 for
            param in params))
        self.assertTrue(any(
            param.name == "min_sim" and param.default_value == 1.0e-6 and param.lower == 1.0e-10 and param.upper ==
            1.0e-2 for param in params))


if __name__ == '__main__':
    unittest.main()


import pandas as pd
from ConfigSpace import Configuration, ConfigurationSpace

# from smac.facade.smac_hpo_facade import SMAC4HPO
# from smac.scenario.scenario import Scenario

from smac import HyperparameterOptimizationFacade
from smac.initial_design import RandomInitialDesign
from smac.intensifier import Intensifier
from smac.scenario import Scenario

from lenskit.data import Dataset, ItemListCollection

from lkauto.explicit.explicit_evaler import ExplicitEvaler
from lkauto.implicit.implicit_evaler import ImplicitEvaler
from lkauto.utils.filer import Filer
from lkauto.utils.get_default_configuration_space import get_default_configuration_space

from typing import Tuple
import logging


def bayesian_optimization(train: Dataset,
                          user_feedback: str,
                          validation: ItemListCollection = None,
                          cs: ConfigurationSpace = None,
                          optimization_metric=None,
                          time_limit_in_sec: int = 2700,
                          num_evaluations: int = 100,
                          random_state=None,
                          split_folds: int = 1,
                          split_strategie: str = 'user_based',
                          split_frac: float = 0.25,
                          ensemble_size: int = 50,
                          num_recommendations: int = 10,
                          minimize_error_metric_val: bool = True,
                          filer: Filer = None) -> Tuple[Configuration, pd.DataFrame]:
    """
        returns the best configuration found by bayesian optimization.
        The bayesian_optimization method will use SMAC3 to find the best
        performing configuration for the given train split.
        The ConfigurationSpace can consist of hyperparameters for a
        single algorithm or a combination of algorithms.

        Parameters
        ----------
        train : pd.DataFrame
            Pandas Dataframe outer train split.
        validation : pd.DataFrame
            Pandas Dataframe validation split.
        cs : ConfigurationSpace
            ConfigurationSpace with all algorithms and hyperparameter ranges defined.
        time_limit_in_sec : int
            time limit in seconds for the optimization process
        num_evaluations : int
            number of samples to be drawn from the ConfigurationSpace
        optimization_metric : function
            LensKit prediction accuracy metric to optimize for.
        minimize_error_metric_val : bool
            Bool that decides if the error metric should be minimized or maximized.
        user_feedback : str
            Defines if the dataset contains explicit or implicit feedback.
        random_state: int
        split_folds : int
            number of folds for cross validation
        split_strategie : str
            cross validation strategie (user_based or item_based)
        split_frac : float
            fraction of the dataset to be used for the validation split. If num_folds > 1, the fraction value
            will be ignored.
        ensemble_size : int
            number of models to be used in the ensemble of rating prediction tasks. This value will be ignored
            for recommender tasks.
        num_recommendations : int
            number of recommendations to be made for each user. This value will be ignored for rating prediction.
        filer : Filer
            filer to manage LensKit-Auto output

        Returns
        -------
        incumbent : Configuration
            best configuration found by bayesian optimization
        top_n_runs : pd.DataFrame
            top n runs found by bayesian optimization
    """
    logger = logging.getLogger('lenskit-auto')
    logger.info('--Start Bayesian Optimization--')

    # get SMAC output directory
    output_dir = filer.get_smac_output_directory_path()

    # initialize Evaler for SMAC evaluations
    if user_feedback == 'explicit':
        evaler = ExplicitEvaler(train=train,
                                validation=validation,
                                optimization_metric=optimization_metric,
                                filer=filer,
                                random_state=random_state,
                                split_folds=split_folds,
                                split_strategy=split_strategie,
                                split_frac=split_frac,
                                ensemble_size=ensemble_size,
                                minimize_error_metric_val=minimize_error_metric_val)
    elif user_feedback == 'implicit':
        evaler = ImplicitEvaler(train=train,
                                validation=validation,
                                optimization_metric=optimization_metric,
                                filer=filer,
                                random_state=random_state,
                                split_folds=split_folds,
                                split_strategy=split_strategie,
                                split_frac=split_frac,
                                num_recommendations=num_recommendations,
                                minimize_error_metric_val=minimize_error_metric_val)
    else:
        raise ValueError('feedback must be either explicit or implicit')

    # get pre-defined ConfiguraitonSpace if none is provided
    if cs is None:
        logger.debug('initializing default ConfigurationSpace')
        cs = get_default_configuration_space(data=train,
                                             validation=validation,
                                             feedback='explicit',
                                             random_state=random_state)

    # initial_configuraition = get_default_configurations(cs)

    scenario = Scenario(configspace=cs,
                        deterministic=True,
                        n_trials=num_evaluations,
                        walltime_limit=time_limit_in_sec,
                        output_directory=output_dir
                        )

    def target_function(config, seed=0, budget=None, instance=None):
        return evaler.evaluate(config)

    smac = HyperparameterOptimizationFacade(
        scenario=scenario,
        target_function=target_function,
        initial_design=RandomInitialDesign(scenario),
        intensifier=Intensifier(scenario),
        overwrite=True
    )

    incumbent = smac.optimize()

    logger.info('--End Bayesian Optimization--')

    # return best model configuration
    if user_feedback == 'explicit':
        # save top n runs
        filer.save_dataframe_as_csv(evaler.top_n_runs, '', 'top_n_runs')
        return incumbent, evaler.top_n_runs
    elif user_feedback == 'implicit':
        return incumbent
    else:
        raise ValueError('feedback must be either explicit or implicit')
